"""
KinsCat
Kinematic Scattering Theory
Copyright by Gerd Duscher

The University of Tennessee, Knoxville
Department of Materials Science & Engineering

Sources:
   Scattering Theory:
   Zuo and Spence, "Advanced TEM", 2017

   Spence and Zuo, Electron Microdiffraction, Plenum 1992

   Atomic Form Factor:
       Kirkland: Advanced Computing in Electron Microscopy 2nd edition
       Appendix C

Units:
    everything is in SI units, except length is given in nm.

Usage:
    See the notebooks for examples of these routines

All the input and output is done through a dictionary
"""

# numerical packages used
import numpy as np
import scipy.constants as const
import itertools
from scipy import spatial

import sidpy
# plotting package used
import matplotlib.pylab as plt  # basic plotting
import matplotlib as mpl
# from mpl_toolkits.mplot3d import Axes3D  # 3D plotting
from matplotlib.patches import Circle  # , Ellipse, Rectangle
from matplotlib.collections import PatchCollection

from ase.io import read  # , write
import pyTEMlib.file_tools as ft
from pyTEMlib.crystal_tools import *
import os


_version_ = 0.5

print('Using KinsCat library version ', _version_, ' by G.Duscher')
_spglib_present = True
try:
    import spglib
except ModuleNotFoundError:
    _spglib_present = False

if _spglib_present:
    print('Symmetry functions of spglib enabled')
else:
    print('spglib not installed; Symmetry functions of spglib disabled')


def read_poscar(file_name=None):
    """
    Open file dialog to select a POSCAR file from VASP
    """
    if file_name is None:
        file_name = ft.openfile_dialog('POSCAR (POSCAR*.txt);;All files (*)')

    # use ase package to read file
    base = os.path.basename(file_name)
    base_name = os.path.splitext(base)[0]
    crystal = read(file_name, format='vasp', parallel=False)

    # make dictionary and plot structure (not essential for further notebook)
    tags = {'unit_cell': crystal.cell * 1e-1, 'elements': crystal.get_chemical_symbols(),
            'base': crystal.get_scaled_positions(), 'max_bond_length': 0.23, 'name': base_name}
    return tags


# Some Structure Determination Routines
def example(verbose=True):
    """
    same as Zuo_fig_3_18
    """
    return Zuo_fig_3_18(verbose=verbose)


def Zuo_fig_3_18(verbose=True):
    """
    Input for Figure 3.18 in Zuo and Spence \"Advanced TEM\", 2017

    This input acts as an example as well as a reference

    Returns:
        dictionary: tags is the dictionary of all input and output paramter needed to reproduce that figure.
    """

    # INPUT
    # Create Silicon structure (Could be produced with Silicon routine)
    if verbose:
        print('Sample Input for Figure 3.18 in Zuo and Spence \"Advanced TEM\", 2017')
    tags = {'crystal_name': 'Silicon'}
    if verbose:
        print('tags[\'crystal\'] = ', tags['crystal_name'])
    a = 0.514  # nm

    tags['lattice_parameter_nm'] = a
    if verbose:
        print('tags[\'lattice_parameter_nm\'] =', tags['lattice_parameter_nm'])
    tags['unit_cell'] = [[a, 0, 0], [0, a, 0], [0, 0, a]]
    if verbose:
        print('tags[\'unit_cell\'] =', tags['unit_cell'])
    tags['elements'] = list(itertools.repeat('Si', 8))
    if verbose:
        print('tags[\'elements\'] =', tags['elements'])
    base = [(0., 0., 0.), (0.5, 0.0, 0.5), (0.5, 0.5, 0.), (0., 0.5, 0.5)]
    tags['base'] = np.array(base + (np.array(base) + (.25, .25, .25)).tolist())
    if verbose:
        print('tags[\'base\'] =', tags['base'])

    # Define Experimental Conditions
    tags['convergence_angle_mrad'] = 7

    tags['acceleration_voltage_V'] = 101.6*1000.0  # V
    if verbose:
        print('tags[\'acceleration_voltage_V\'] =', tags['acceleration_voltage_V'])

    tags['convergence_angle_mrad'] = 7.1  # mrad;  0 is parallel illumination
    if verbose:
        print('tags[\'convergence_angle_mrad\'] =', tags['convergence_angle_mrad'])

    tags['zone_hkl'] = np.array([-2, 2, 1])  # incident neares zone axis: defines Laue Zones!!!!
    if verbose:
        print('tags[\'zone_hkl\'] =', tags['zone_hkl'])
    tags['mistilt'] = np.array([0, 0, 0])  # mistilt in degrees
    if verbose:
        print('tags[\'mistilt\'] =', tags['mistilt'])

    # Define Simulation Parameters

    tags['Sg_max'] = .2  # 1/nm  maximum allowed excitation error
    if verbose:
        print('tags[\'Sg_max\'] =', tags['Sg_max'])

    tags['hkl_max'] = 9   # Highest evaluated Miller indices
    if verbose:
        print('tags[\'hkl_max\'] =', tags['hkl_max'])

        print('##################')
        print('# Output Options #')
        print('##################')

    # Output options

    tags['background'] = 'black'  # 'white'  'grey'
    if verbose:
        print('tags[\'background\'] =', tags['background'], '# \'white\',  \'grey\' ')
    tags['color map'] = 'plasma'
    if verbose:
        print('tags[\'color map\'] =', tags['color map'], '#,\'cubehelix\',\'Greys\',\'jet\' ')

    tags['plot HOLZ'] = 1
    if verbose:
        print('tags[\'plot HOLZ\'] =', tags['plot HOLZ'])
    tags['plot HOLZ excess'] = 1
    if verbose:
        print('tags[\'plot HOLZ excess\'] =', tags['plot HOLZ excess'])
    tags['plot Kikuchi'] = 1
    if verbose:
        print('tags[\'plot Kikuchi\'] =', tags['plot Kikuchi'])
    tags['plot reflections'] = 1
    if verbose:
        print('tags[\'plot reflections\'] =', tags['plot reflections'])

    tags['label HOLZ'] = 0
    if verbose:
        print('tags[\'label HOLZ\'] =', tags['label HOLZ'])
    tags['label Kikuchi'] = 0
    if verbose:
        print('tags[\'label Kikuchi\'] =', tags['label Kikuchi'])
    tags['label reflections'] = 0
    if verbose:
        print('tags[\'label reflections\'] =', tags['label reflections'])

    tags['label color'] = 'black'
    if verbose:
        print('tags[\'label color\'] =', tags['label color'])
    tags['label size'] = 10
    if verbose:
        print('tags[\'label size\'] =', tags['label size'])

    tags['color Laue Zones'] = ['red',  'blue', 'green', 'blue', 'green']   # , 'green', 'red'] #for OLZ give a sequence
    if verbose:
        print('tags[\'color Laue Zones\'] =', tags['color Laue Zones'], ' #[\'red\', \'blue\', \'lightblue\']')

    tags['color Kikuchi'] = 'green'
    if verbose:
        print('tags[\'color Kikuchi\'] =', tags['color Kikuchi'])
    tags['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    if verbose:
        print('tags[\'linewidth HOLZ\'] =', tags['linewidth HOLZ'], '# -1: linewidth according to intensity '
                                                                    '(structure factor F^2)')
    tags['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2
    if verbose:
        print('tags[\'linewidth Kikuchi\'] =', tags['linewidth Kikuchi'], '# -1: linewidth according to intensity '
                                                                          '(structure factor F^2)')

    tags['color reflections'] = 'intensity'  # 'Laue Zone'
    if verbose:
        print('tags[\'color reflections\'] =', tags['color reflections'], '#\'Laue Zone\' ')
    tags['color zero'] = 'white'  # 'None', 'white', 'blue'
    if verbose:
        print('tags[\'color zero\'] =', tags['color zero'], '#\'None\', \'white\', \'blue\' ')
    tags['color ring zero'] = 'None'  # 'Red' #'white' #, 'None'
    if verbose:
        print('tags[\'color ring zero\'] =', tags['color ring zero'], '#\'None\', \'white\', \'Red\' ')

        print('########################')
        print('# End of Example Input #')
        print('########################\n\n')
    return tags


def zone_mistilt(zone, angles):
    """ Rotation of zone axis by mistilt

    Parameters
    ----------
    zone: list or numpy array of int
        zone axis in Miller indices
    angles: ist or numpy array of float
        list of mistilt angles in degree

    Returns
    -------
    new_zone_axis: np.ndarray (3)
        new tilted zone axis

    """
    if not isinstance(angles, (np.ndarray, list)):
        raise TypeError('angles must be a list of float of length 3')
    if len(angles) != 3:
        raise TypeError('angles must be a list of float of length 3')
    if not isinstance(zone, (np.ndarray, list)):
        raise TypeError('Miller indices must be a list of int of length 3')
    alpha, beta, gamma = np.radians(angles)

    # first we rotate alpha about x axis
    c, s = np.cos(alpha), np.sin(alpha)
    rotx = np.array([[1, 0, 0], [0, c, -s], [0, s, c]])

    # second we rotate beta about y axis
    c, s = np.cos(beta), np.sin(beta)
    roty = np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])

    # third we rotate gamma about z-axis
    c, s = np.cos(gamma), np.sin(gamma)
    rotz = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])

    return np.dot(np.dot(np.dot(zone, rotx), roty), rotz)


def get_symmetry(unit_cell, base, atoms, verbose=True):
    """
    Symmetry analysis with spglib

    spglib must be installed
    """
    if _spglib_present:
        if verbose:
            print('#####################')
            print('# Symmetry Analysis #')
            print('#####################')

        atomic_number = []
        for i in range(len(atoms)):
            a = atoms[i]
            b = base[i]
            atomic_number.append(electronFF[a]['Z'])
            if verbose:
                print(f'{i+1}: {atomic_number[i]} = {2} : [{base[i][0]:.2f}, {base[i][1]:.2f}, {base[i][2]:.2f}]')

        lattice = (unit_cell, base, atomic_number)
        spgroup = spglib.get_spacegroup(lattice)
        sym = spglib.get_symmetry(lattice)

        if verbose:
            print("  Spacegroup  is %s." % spgroup)
            print('  Crystal has {0} symmetry operation'.format(sym['rotations'].shape[0]))

        p_lattice, p_positions, p_numbers = spglib.find_primitive(lattice, symprec=1e-5)
        print("\n########################\n #Basis vectors of primitive Cell:")
        for i in range(3):
            print('[{0:.4f}, {1:.4f}, {2:.4f}]'.format(p_lattice[i][0], p_lattice[i][1], p_lattice[i][2]))

        print('There {0} atoms and {1} species in primitive unit cell:'.format(len(p_positions), p_numbers))
    else:
        print('spglib is not installed')

    return True


def ball_and_stick(tags, extend=1, max_bond_length=0.):
    """Calculates the data to plot a ball and stick model

    Parameters
    ----------
    tags: dict
        dictionary containing the 'unit_cell', 'base' and 'elements' tags.

    extend: 1 or 3 integers
        The *extend* argument scales the effective cell in which atoms
        will be included. It must either be a list of three integers or a single
        integer scaling all 3 directions.  By setting this value to one,
        all  corner and edge atoms will be included in the returned cell.
        This will of cause make the returned cell non-repeatable, but this is
        very useful for visualisation.

    max_bond_length: 1 float
        The max_bond_length argument defines the distance for which a bond will be shown.
        If max_bond_length is zero, the tabulated atom radii will be used.

    Returns
    -------
    corners,balls, Z, bonds: lists
        These lists can be used to plot the
    unit cell:
        for x, y, z in corners:
            l=mlab.plot3d( x,y,z, tube_radius=0.002)
    bonds:
        for x, y, z in bonds:
            mlab.plot3d( x,y,z, tube_radius=0.02)
    and atoms:
        for i,atom in enumerate(balls):
            mlab.points3d(atom[0],atom[1],atom[2],
                          scale_factor=0.1,##ks.vdw_radii[Z[i]]/5,
                          resolution=20,
                          color=tuple(ks.jmol_colors [Z[i]]),
                          scale_mode='none')

    Please note that you'll need the *Z* list for coloring, or for radii that depend on elements
    """

    # Check in which form extend is given
    if isinstance(extend, int):
        extend = [extend]*3

    extend = np.array(extend, dtype=int)

    # Make the x,y, and z multiplicators
    if len(extend) == 3:
        x = np.linspace(0, extend[0], extend[0]+1)
        y = np.linspace(0, extend[1], extend[1]+1)
        z = np.linspace(0, extend[2], extend[2]+1)
    else:
        print('wrong extend parameter')
        return

    # Check whether this is the right kind of dictionary
    if 'unit_cell' not in tags:
        return
    cell = tags['unit_cell']

    # Corners and Outline of unit cell
    h = (0, 1)
    corner_vectors = np.dot(np.array(list(itertools.product(h, h, h))), cell)
    trace = [[0, 1], [1, 3], [3, 2], [2, 0], [0, 4], [4, 5], [5, 7], [6, 7], [6, 4], [1, 5], [2, 6], [3, 7]]
    corners = []
    for s, e in trace:
        corners.append([*zip(corner_vectors[s], corner_vectors[e])])

    # ball position and elements in supercell
    super_cell = np.array(list(itertools.product(x, y, z)))  # all evaluated Miller indices

    pos = np.add(super_cell, np.array(tags['base'])[:, np.newaxis])

    atomic_number = []
    for i in range(len(tags['elements'])):
        atomic_number.append(electronFF[tags['elements'][i]]['Z'])

    # List of bond lengths taken from electronFF database below
    bond_lengths = []
    for atom in tags['elements']:
        bond_lengths.append(electronFF[atom]['bond_length'][0])

    # extend list of atomic numbers
    zpp = []
    for z in atomic_number:
        zpp.append([z]*pos.shape[1])
    zpp = np.array(zpp).flatten()

    # reshape supercell atom positions
    pos = pos.reshape((pos.shape[0]*pos.shape[1], pos.shape[2]))

    # Make a mask that excludes all atoms outside of super cell
    maskz = pos[:, 2] <= extend[2]
    masky = pos[:, 1] <= extend[1]
    maskx = pos[:, 0] <= extend[0]
    mask = np.logical_and(maskx, np.logical_and(masky, maskz))   # , axis=1)

    # Only use balls and elements inside super cell
    balls = np.dot(pos[mask], cell)
    atomic_number = zpp[mask]

    # Get maximum bond length from list of bond -lengths taken from electronFF database
    if max_bond_length == 0:
        max_bond_length = np.median(bond_lengths)/5
    # Check nearest neighbours within max_bond_length
    tree = spatial.KDTree(balls)
    nearest_neighbours = np.array(tree.query(balls, k=8, distance_upper_bound=max_bond_length))

    # Calculate bonds
    bonds = []
    bond_indices = []
    for i in range(nearest_neighbours.shape[1]):
        for j in range(nearest_neighbours.shape[2]):
            if nearest_neighbours[0, i, j] < max_bond_length:
                if nearest_neighbours[0, i, j] > 0:
                    # print(atoms[i],atoms[int(bonds[1,i,j])],bonds[:,i,j])
                    bonds.append([*zip(balls[i], balls[int(nearest_neighbours[1, i, j])])])
                    bond_indices.append([i, int(nearest_neighbours[1, i, j])])
    return corners, balls, atomic_number, bonds


def plot_unitcell_mayavi(tags):
    """Makes a 3D plot of crystal structure

    Parameters
    ----------
    tags: dict
        Dictionary with tags: 'unit_cell, 'elements', 'base'

    Returns
    -------
    3D plot

    Dependencies
    ------------
    ball_and_stick function of KinsCat
    mlab of mayavi
    """

    from mayavi import mlab

    # Make sure "max_bond_length" and "extend" variables are initialized
    if 'max_bond_length' not in tags:
        max_bond_length = 0.
    else:
        max_bond_length = tags['max_bond_length']

    if 'extend' not in tags:
        extend = 1
    else:
        extend = tags['extend']

    # get balls, sticks and atomic numbers for colors and sizes
    corners, balls, atomic_number, bonds = ball_and_stick(tags, extend=extend, max_bond_length=max_bond_length)

    print('Now plotting')
    fig = mlab.figure(1, bgcolor=(0, 0, 0), size=(350, 350))

    mlab.clf()  # clear figure

    # parallel projection
    mlab.gcf().scene.parallel_projection = True
    mlab.gcf().scene.camera.parallel_scale = 5

    # plot unit cell
    for x, y, z in corners:
        ll = mlab.plot3d(x, y, z, tube_radius=0.002)

    # plot bonds as sticks
    for x, y, z in bonds:
        mlab.plot3d(x, y, z, tube_radius=0.02)

    # plot atoms
    for i, atom in enumerate(balls):
        mlab.points3d(atom[0], atom[1], atom[2],
                      scale_factor=0.1,  # ks.vdw_radii[Z[i]]/50,
                      resolution=20,
                      color=tuple(jmol_colors[atomic_number[i]]),
                      scale_mode='none')

    # parallel projection
    mlab.gcf().scene.parallel_projection = True
    mlab.gcf().scene.camera.parallel_scale = 5
    # show plot
    mlab.show()


def plot_unitcell(tags):
    """
    Simple plot of unit cell
    """

    if 'max_bond_length' not in tags:
        max_bond_length = 0.
    else:
        max_bond_length = tags['max_bond_length']

    if 'extend' not in tags:
        extend = 1
    else:
        extend = tags['extend']

    corners, balls, atomic_number, bonds = ball_and_stick(tags, extend=extend, max_bond_length=max_bond_length)

    maximum_position = balls.max()*1.05
    maximum_x = balls[:, 0].max()
    maximum_y = balls[:, 1].max()
    maximum_z = balls[:, 2].max()

    balls = balls - [maximum_x/2, maximum_y/2, maximum_z/2]

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    # draw unit_cell
    for x, y, z in corners:
        ax.plot3D(x-maximum_x/2, y-maximum_y/2, z-maximum_z/2, color="blue")

    # draw bonds
    for x, y, z in bonds:
        ax.plot3D(x-maximum_x/2, y-maximum_y/2, z-maximum_z/2, color="black", linewidth=4)  # , tube_radius=0.02)

    # draw atoms
    for i, atom in enumerate(balls):
        ax.scatter(atom[0], atom[1], atom[2],
                   color=tuple(jmol_colors[atomic_number[i]]),
                   alpha=1.0, s=50)
    maximum_position = balls.max()*1.05
    ax.set_proj_type('ortho')

    ax.set_zlim(-maximum_position/2, maximum_position/2)
    ax.set_ylim(-maximum_position/2, maximum_position/2)
    ax.set_xlim(-maximum_position/2, maximum_position/2)

    if 'name' in tags:
        ax.set_title(tags['name'])

    ax.set_xlabel('x [nm]')
    ax.set_ylabel('y [nm]')
    ax.set_zlabel('z [nm]')


# The metric tensor of the lattice.
def metric_tensor(matrix):
    """
    The metric tensor of the lattice.

    Usage:
        metric_tensor(unit_cell)
    """
    metric_tensor2 = np.dot(matrix, matrix.T)
    return metric_tensor2


def vector_norm(g):
    """ Length of vector

    depreciated - use np.linalg.norm
    """
    g = np.array(g)
    return np.sqrt(g[:, 0]**2+g[:, 1]**2+g[:, 2]**2)


def make_pretty_labels(hkls, hex_label=False):
    """Make pretty labels

    Parameters
    ----------
    hkls: np.ndarray
        a numpy array with all the Miller indices to be labeled
    hex_label: boolean - optional
        if True this will make for Miller indices.

    Returns
    -------
    hkl_label: list
        list of labels in Latex format
    """
    hkl_label = []
    for i in range(len(hkls)):
        h, k, l = np.array(hkls)[i]

        if h < 0:
            h_string = r'[$\bar {'+str(int(-h))+'},'
        else:
            h_string = r'[$\bar {'+str(int(h))+'},'
        if k < 0:
            k_string = r'\bar {'+str(int(-k))+'},'
        else:
            k_string = str(int(k))+','
        if hex_label:
            ii = -(h+k)
            if ii < 0:
                k_string = k_string + r'\bar {'+str(int(-ii))+'},'
            else:
                k_string = k_string + str(int(ii))+','
        if l < 0:
            l_string = r'\bar {'+str(int(-l))+'} $]'
        else:
            l_string = str(int(l))+'} $]'
        label = h_string+k_string+l_string
        hkl_label.append(label)
    return hkl_label


def get_wavelength(e0):
    """
    Calculates the relativistic corrected de Broglie wavelength of an electron in nm

    Input:
    ------
        acceleration voltage in volt
    Output:
    -------
        wave length in nm
    """
    if not isinstance(e0, (int, float)):
        raise TypeError('Acceleration voltage has to be a real number')
    eV = const.e * e0
    return const.h/np.sqrt(2*const.m_e*eV*(1+eV/(2*const.m_e*const.c**2)))*10**9


# Determine rotation matrix to tilt zone axis onto z-axis
# We determine spherical coordinates to do that
def get_rotation_matrix(zone, verbose=False):
    """Calculates the rotation matrix to rotate the zone axis parallel to the cartasian z-axis.

     We use spherical coordinates to first rotate around the z-axis and then around the y-axis.
     This makes it easier to apply additional tilts, than to use the cross product to determine a single rotation
     axis (Rodrigues Formula)

     We start from the dot product of zone axis  and unit cell will accomplish that.


     Parameters
     ----------
     zone: list of int or np.ndarray of length 3
        axis has to be in cartesian coordinates.

    Returns
    -------
    rotation_matrix: np.ndarray(3,3)
    theta: float (degrees)
    phi: float (degrees)
    """

    # spherical coordinates of zone
    zone = np.array(zone)
    r = np.sqrt((zone*zone).sum())
    theta = np.arccos(zone[2]/r)
    if zone[0] < 0:
        theta = -theta
    if zone[0] == 0:
        phi = np.pi/2
    else:
        phi = (np.arctan(zone[1]/zone[0]))

    if verbose:
        print('Rotation theta ', np.degrees(theta), ' phi ', np.degrees(phi))
    # unit = np.array([[1, 0, 0],[0,1, 0],[0, 0,1]])

    # first we rotate phi about z-axis
    c, s = np.cos(phi), np.sin(phi)
    rotz = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])

    # second we rotate theta about y axis
    c, s = np.cos(theta), np.sin(theta)
    roty = np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])

    # the rotation now makes zone-axis coincide with plane normal
    return np.dot(rotz, roty), np.degrees(theta), np.degrees(phi)


def check_sanity(tags):
    """
    Check sanity of input parameters
    """
    stop = False
    if 'unit_cell' not in tags:
        print(' No unit_cell defined')
        stop = True
    elif 'base' not in tags:
        print(' No base defined')
        stop = True
    elif 'elements' not in tags:
        print(' No atoms defined')
        stop = True
    elif 'acceleration_voltage_V' not in tags:
        print(' No acceleration_voltage_V defined')
        stop = True
    elif 'zone_hkl' not in tags:
        print(' No zone_hkl defined')
        stop = True
    elif 'Sg_max' not in tags:
        print(' No Sg_max defined')
        stop = True
    elif 'hkl_max' not in tags:
        print(' No hkl_max defined')
        stop = True

    elif 'crystal_name' not in tags:
        tags['crystal_name'] = 'undefined'
        print('tags[\'crystal\'] = \'undefined\'')
    elif 'mistilt' not in tags:
        tags['mistilt'] = [0., 0., 0.]
        print('tags[\'mistilt\'] = [0., 0., 0.]')
    elif 'convergence_angle_mrad' not in tags:
        tags['convergence_angle_mrad'] = 0.
        print('tags[\'convergence_angle_mrad\'] = 0')

    return not stop


def ring_pattern_calculation(tags, verbose=False):
    """
    Calculate the ring diffraction pattern of a crystal structure

    Parameters
    ----------
    tags: dict
        dictionary of crystal structure

    Returns
    -------
    tags: dict
        dictionary with diffraction information added

    """
    # Reciprocal Lattice
    # We use the linear algebra package of numpy to invert the unit_cell "matrix"
    reciprocal_unit_cell = np.linalg.inv(tags['unit_cell']).T  # transposed of inverted unit_cell

    # INPUT
    hkl_max = 7   # maximum allowed Miller index

    acceleration_voltage = 200.0*1000.0  # V
    wave_length = get_wavelength(acceleration_voltage)

    h = np.linspace(-hkl_max, hkl_max, 2*hkl_max+1)    # all to be evaluated single Miller Index
    hkl = np.array(list(itertools.product(h, h, h)))  # all to be evaluated Miller indices
    g_hkl = np.dot(hkl, reciprocal_unit_cell)

    # Calculate Structure Factors

    structure_factors = []

    base = np.dot(tags['base'], tags['unit_cell'])  # transformation from relative to Cartesian coordinates
    for j in range(len(g_hkl)):
        F = 0
        for b in range(len(base)):
            # Atomic form factor for element and momentum change (g vector)
            f = feq(tags['elements'][b], np.linalg.norm(g_hkl[j]))
            F += f * np.exp(-2*np.pi*1j*(g_hkl[j]*base[b]).sum())
        structure_factors.append(F)
    F = structure_factors = np.array(structure_factors)

    # Allowed reflections have a non zero structure factor F (with a  bit of numerical error)
    allowed = np.absolute(structure_factors) > 0.001

    distances = np.linalg.norm(g_hkl, axis=1)

    if verbose:
        print(f' Of the evaluated {hkl.shape[0]} Miller indices {allowed.sum()} are allowed. ')
    # We select now all the
    zero = distances == 0.
    allowed = np.logical_and(allowed, np.logical_not(zero))

    F = F[allowed]
    g_hkl = g_hkl[allowed]
    hkl = hkl[allowed]
    distances = distances[allowed]

    sorted_allowed = np.argsort(distances)

    distances = distances[sorted_allowed]
    hkl = hkl[sorted_allowed]
    F = F[sorted_allowed]

    # How many have unique distances and what is their multiplicity
    unique, indices = np.unique(distances, return_index=True)

    if verbose:
        print(f' Of the {allowed.sum()} allowed Bragg reflections there are {len(unique)} families of reflections.')

    intensity = np.absolute(F[indices]**2*(np.roll(indices, -1)-indices))
    if verbose:
        print('\n index \t  hkl \t      1/d [1/nm]       d [pm]     F     multip.  intensity')
    family = []
    out_tags = {}
    for j in range(len(unique)-1):
        i = indices[j]
        i2 = indices[j+1]
        family.append(hkl[i+np.argmax(hkl[i:i2].sum(axis=1))])
        index = '{'+f'{family[j][0]:.0f} {family[j][1]:.0f} {family[j][2]:.0f}'+'}'
        if verbose:
            print(f'{i:3g}\t {index} \t  {distances[i]:.2f}  \t {1/distances[i]*1000:.0f} \t {np.absolute(F[i]):.2f},'
                  f' \t  {indices[j+1]-indices[j]:3g} \t {intensity[j]:.2f}')
        out_tags[index] = {}
        out_tags[index]['reciprocal_distance'] = distances[i]
        out_tags[index]['real_distance'] = 1/distances[i]
        out_tags[index]['F'] = np.absolute(F[i])
        out_tags[index]['multiplicity'] = indices[j+1]-indices[j]
        out_tags[index]['intensity'] = intensity[j]
    return out_tags


def kinematic_scattering(tags, verbose=False):
    """
        All kinematic scattering calculation

        Calculates Bragg spots, Kikuchi lines, excess, and deficient HOLZ lines

        Parameters
        ----------
        tags: dict
            dictionary with crystal structure:
            'unit_cell', 'base' 'elements'
            and with experimental parameters:
            'acceleration_voltage_V', 'zone_hkl', 'Sg_max', 'hkl_max'
            Optional parameters are:
            'mistilt', convergence_angle_mrad', and 'crystal_name'
        verbose:  boolean
            True will give extended output of the calculation

        Returns
        -------
        dict:
            There are three sub_dictionaries:
            ['allowed'], ['forbidden'], and ['HOLZ']
            ['allowed'] and ['forbidden'] dictionaries contain:
                ['Sg'], ['hkl'], ['g'], ['structure factor'], ['intensities'],
                ['ZOLZ'], ['FOLZ'], ['SOLZ'], ['HOLZ'], ['HHOLZ'], ['label'], and ['Laue_zone']
            the ['HOLZ'] dictionary contains:
                ['slope'], ['distance'], ['theta'], ['g deficient'], ['g excess'], ['hkl'], ['intensities'],
                ['ZOLZ'], ['FOLZ'], ['SOLZ'], ['HOLZ'], and  ['HHOLZ']
            Please note that the Kikuchi lines are the HOLZ lines of ZOLZ

            There are also a few parameters stored in the main dictionary:
                ['wave_length_nm'], ['reciprocal_unit_cell'], ['inner_potential_V'], ['incident_wave_vector'],
                ['volume'], ['theta'], ['phi'], and ['incident_wave_vector_vacuum']
    """

    if not check_sanity(tags):
        print('Input is not complete, stopping')
        print('Try \'example()\' for example input')
        return

    tags['wave_length_nm'] = get_wavelength(tags['acceleration_voltage_V'])

    # ###########################################
    # reciprocal_unit_cell
    # ###########################################
    unit_cell = np.array(tags['unit_cell'])
    # We use the linear algebra package of numpy to invert the unit_cell "matrix"
    reciprocal_unit_cell = np.linalg.inv(unit_cell).T  # transposed of inverted unit_cell
    tags['reciprocal_unit_cell'] = reciprocal_unit_cell

    if verbose:
        print('reciprocal_unit_cell')
        print(np.round(reciprocal_unit_cell, 3))

    # ###########################################
    # Incident wave vector K0 in vacuum and material
    # ###########################################

    # Incident wave vector K0 in vacuum and material
    U0 = 0
    for i in range(len(tags['elements'])):
        a = tags['elements'][i]
        U0 += feq(a, 0)*0.023933754

    tags['volume'] = np.linalg.det(tags['unit_cell'])
    volume = tags['volume']*1000  # Needs to be in Angstrom for form factors

    AngstromConversion = 1.0e10  # So [1A (in m)] * AngstromConversion = 1
    NanometerConversion = 1.0e9

    ScattFacToVolts = (const.h**2)*(AngstromConversion**2)/(2*np.pi*const.m_e*const.e)*volume
    U0 = U0*ScattFacToVolts
    tags['inner_potential_A'] = U0
    tags['inner_potential_V'] = U0*ScattFacToVolts
    if verbose:
        print('The inner potential is {0:.3f}kV'.format(tags['inner_potential_V']/1000))

    # Calculating incident wave vector magnitude 'K0' in material
    wl = tags['wave_length_nm']
    tags['incident_wave_vector_vacuum'] = 1/wl

    K0 = tags['incident_wave_vector'] = np.sqrt(1/wl**2 - (U0/volume*100.))  # 1/nm

    tags['convergence_angle_nm-1'] = K0*np.sin(tags['convergence_angle_mrad']/1000.)

    if verbose:
        print('Magnitude of incident wave vector in material {0:.1f} 1/nm and vacuum {1:.1f} 1/nm'.format(K0, 1/wl))
        print('The convergence angle of {0}mrad = {1:.2f} 1/nm'.format(tags['convergence_angle_mrad'],
                                                                       tags['convergence_angle_nm-1']))

    # ############
    # Rotate
    # ############

    # first we take care of mistilt: zone axis is then in fractional Miller indices
    zone = tags['zone'] = zone_mistilt(tags['zone_hkl'], tags['mistilt'])

    # zone axis in global coordinate system
    zone_vector = np.dot(zone, reciprocal_unit_cell)

    rotation_matrix, theta, phi = get_rotation_matrix(zone_vector, verbose=False)

    if verbose:
        print('Rotation angles are {0:.1f} deg and {1:.1f} deg'.format(theta, phi))
    tags['theta'] = theta
    tags['phi'] = phi

    # rotate incident wave vector
    w_vector = np.dot(zone_vector, rotation_matrix)
    K0_unit_vector = w_vector / np.linalg.norm(w_vector)  # incident unit wave vector
    K0_vector = K0_unit_vector*K0                         # incident  wave vector

    if verbose:
        print('Center of Ewald sphere ', K0_vector)

    # #######################
    # Find all Miller indices whose reciprocal point lays near the Ewald sphere with radius K0
    # within a maximum excitation error Sg
    # #######################

    hkl_max = tags['hkl_max']
    Sg_max = tags['Sg_max']  # 1/nm  maximum allowed excitation error

    h = np.linspace(-hkl_max, hkl_max, 2*hkl_max+1)  # all evaluated single Miller Indices
    hkl = np.array(list(itertools.product(h, h, h)))  # all evaluated Miller indices
    g = np.dot(hkl, reciprocal_unit_cell)  # all evaluated reciprocal_unit_cell points
    g_norm = np.linalg.norm(g, axis=1)   # length of all vectors
    not_zero = g_norm > 0
    g = g[not_zero]  # zero reflection will make problems further on, so we exclude it.
    g_norm = g_norm[not_zero]
    hkl = hkl[not_zero]
    g_non_rot = g
    g = np.dot(g, rotation_matrix)

    # #######################
    # Calculate excitation errors for all reciprocal_unit_cell points
    # #######################

    # Zuo and Spence, 'Adv TEM', 2017 -- Eq 3:14
    # S=(K0**2-np.linalg.norm(g - K0_vector, axis=1)**2)/(2*K0)
    gMz = g - K0_vector

    in_sqrt = gMz[:, 2]**2 + np.linalg.norm(gMz, axis=1)**2 - K0**2
    in_sqrt[in_sqrt < 0] = 0.
    S = -gMz[:, 2] - np.sqrt(in_sqrt)

    # #######################
    # Determine reciprocal_unit_cell points with excitation error less than the maximum allowed one: Sg_max
    # #######################

    reflections = abs(S) < Sg_max   # This is now a boolean array with True for all possible reflections

    Sg = S[reflections]
    g_hkl = g[reflections]
    g_hkl_non_rot = g_non_rot[reflections]
    hkl = hkl[reflections]
    g_norm = g_norm[reflections]

    if verbose:
        print('Of the {0} tested reciprocal_unit_cell points, {1} have an excitation error less than {2:.2f} 1/nm'.
              format(len(g), len(g_hkl), Sg_max))

    # #################################
    # Calculate Structure Factors
    # ################################

    structure_factors = []
    """for j  in range(len(g_hkl)):
        F = 0
        for b in range(len(tags['base'])):
            f = feq(tags['elements'][b],np.linalg.norm(g_hkl[j]))
            #F += f * np.exp(-2*np.pi*1j*(hkl*tags['base'][b]).sum()) # may only work for cubic Gerd
            F += f * np.exp(-2*np.pi*1j*(g_hkl_non_rot*np.dot(tags['base'][b],unit_cell)).sum())


        structure_factors.append(F)

    F = structure_factors = np.array(structure_factors)
    """
    base = np.dot(tags['base'], tags['unit_cell'])  # transformation from relative to Cartesian coordinates
    for j in range(len(g_hkl)):
        F = 0
        for b in range(len(base)):
            f = feq(tags['elements'][b], g_norm[j])  # Atomic form factor for element and momentum change (g vector)
            F += f * np.exp(-2*np.pi*1j*(g_hkl_non_rot[j]*base[b]).sum())
        structure_factors.append(F)
    F = structure_factors = np.array(structure_factors)

    # ###########################################
    # Sort reflection in allowed and forbidden #
    # ###########################################

    allowed = np.absolute(F) > 0.000001    # allowed within numerical error

    if verbose:
        print('Of the {0} possible reflection {1} are allowed.'.format(hkl.shape[0], allowed.sum()))

    # information of allowed reflections
    Sg_allowed = Sg[allowed]
    hkl_allowed = hkl[allowed][:]
    g_allowed = g_hkl[allowed, :]
    F_allowed = F[allowed]
    g_norm_allowed = g_norm[allowed]

    tags['allowed'] = {}
    tags['allowed']['Sg'] = Sg_allowed
    tags['allowed']['hkl'] = hkl_allowed
    tags['allowed']['g'] = g_allowed
    tags['allowed']['structure factor'] = F_allowed

    # information of forbidden reflections
    forbidden = np.logical_not(allowed)
    Sg_forbidden = Sg[forbidden]
    hkl_forbidden = hkl[forbidden]
    g_forbidden = g_hkl[forbidden]

    tags['forbidden'] = {}
    tags['forbidden']['Sg'] = Sg_forbidden
    tags['forbidden']['hkl'] = hkl_forbidden
    tags['forbidden']['g'] = g_forbidden

    # ##########################
    # Make pretty labels
    # ##########################
    hkl_label = make_pretty_labels(hkl_allowed)
    tags['allowed']['label'] = hkl_label

    # hkl_label = make_pretty_labels(hkl_forbidden)
    # tags['forbidden']['label'] = hkl_label

    # ###########################
    # Calculate Intensities (of allowed reflections)
    # ###########################

    intensities = np.absolute(F_allowed)**2

    tags['allowed']['intensities'] = intensities

    # ###########################
    # Calculate Laue Zones (of allowed reflections)
    # ###########################
    # Below is the expression given in most books.
    # However, that would only work for orthogonal crystal systems
    # Laue_Zone = abs(np.dot(hkl_allowed,tags['zone_hkl']))  # works only for orthogonal systems

    # This expression works for all crystal systems
    # Remember we have already tilted, and so the dot product is trivial and gives only the z-component.
    length_zone_axis = np.linalg.norm(np.dot(tags['zone_hkl'], tags['unit_cell']))
    Laue_Zone = abs(np.floor(g_allowed[:, 2]*length_zone_axis+0.5))

    tags['allowed']['Laue_Zone'] = Laue_Zone

    ZOLZ_forbidden = abs(np.floor(g_forbidden[:, 2]*length_zone_axis+0.5)) == 0

    tags['forbidden']['Laue_Zone'] = ZOLZ_forbidden
    ZOLZ = Laue_Zone == 0
    FOLZ = Laue_Zone == 1
    SOLZ = Laue_Zone == 2
    HOLZ = Laue_Zone > 0
    HOLZp = Laue_Zone > 2

    tags['allowed']['ZOLZ'] = ZOLZ
    tags['allowed']['FOLZ'] = FOLZ
    tags['allowed']['SOLZ'] = SOLZ
    tags['allowed']['HOLZ'] = HOLZ
    tags['allowed']['HOLZ_plus'] = tags['allowed']['HHOLZ'] = HOLZp

    if verbose:
        print(' There are {0} allowed reflections in the zero order Laue Zone'.format(ZOLZ.sum()))
        print(' There are {0} allowed reflections in the first order Laue Zone'.format((Laue_Zone == 1).sum()))
        print(' There are {0} allowed reflections in the second order Laue Zone'.format((Laue_Zone == 2).sum()))
        print(' There are {0} allowed reflections in the other higher order Laue Zones'.format((Laue_Zone > 2).sum()))

    if verbose == 2:
        print(' hkl  \t Laue zone \t Intensity (*1 and \t log) \t length \n')
        for i in range(len(hkl_allowed)):
            print(' {0} \t {1} \t {2:.3f} \t  {3:.3f} \t  {4:.3f}   '.format(hkl_allowed[i], g_allowed[i],
                                                                             intensities[i], np.log(intensities[i]+1),
                                                                             g_norm_allowed[i]))

    # ##########################
    # Dynamically Activated forbidden reflections
    # ##########################

    double_diffraction = (np.sum(np.array(list(itertools.combinations(hkl_allowed[ZOLZ], 2))), axis=1))

    dynamical_allowed = []
    still_forbidden = []
    for i, hkl in enumerate(hkl_forbidden):
        if ZOLZ_forbidden[i]:
            if hkl.tolist() in double_diffraction.tolist():
                dynamical_allowed.append(i)
            else:
                still_forbidden.append(i)
    tags['forbidden']['dynamically_activated'] = dynamical_allowed
    tags['forbidden']['forbidden'] = dynamical_allowed
    if verbose:
        print('Length of zone axis vector in real space {0} nm'.format(np.round(length_zone_axis, 3)))
        print(f'There are {len(dynamical_allowed)} forbidden but dynamical activated diffraction spots:')
        # print(tags['forbidden']['hkl'][dynamical_allowed])

    # ###################################
    # Calculate HOLZ and Kikuchi Lines #
    # ###################################

    # Dynamic Correction

    # Equation Spence+Zuo 3.86a
    gamma_1 = - 1./(2.*K0) * (intensities / (2.*K0*Sg_allowed)).sum()
    # print('gamma_1',gamma_1)

    # Equation Spence+Zuo 3.84
    Kg = K0 - K0*gamma_1/(g_allowed[:, 2]+1e-15)
    Kg[ZOLZ] = K0

    # print(Kg, Kg.shape)

    # Calculate angle between K0 and deficient cone vector
    # For dynamic calculations K0 is replaced by Kg
    Kg[:] = K0
    dtheta = np.arcsin(g_norm_allowed/Kg/2.)-np.arcsin(np.abs(g_allowed[:, 2])/g_norm_allowed)

    # calculate length of distance of deficient cone to K0 in ZOLZ plane
    gd_length = 2*np.sin(dtheta/2)*K0

    # Calculate nearest point of HOLZ and Kikuchi lines
    g_closest = g_allowed.copy()
    g_closest = g_closest*(gd_length/np.linalg.norm(g_closest, axis=1))[:, np.newaxis]

    g_closest[:, 2] = 0.

    # calculate and save line in Hough space coordinates (distance and theta)
    slope = g_closest[:, 0]/(g_closest[:, 1]+1e-10)
    distance = gd_length
    theta = np.arctan2(g_allowed[:, 0], g_allowed[:, 1])

    tags['HOLZ'] = {}
    tags['HOLZ']['slope'] = slope
    # a line is now given by

    tags['HOLZ']['distance'] = distance
    tags['HOLZ']['theta'] = theta

    tags['HOLZ']['g deficient'] = g_closest
    tags['HOLZ']['g excess'] = g_closest+g_allowed

    tags['HOLZ']['ZOLZ'] = ZOLZ
    tags['HOLZ']['HOLZ'] = HOLZ
    tags['HOLZ']['FOLZ'] = FOLZ
    tags['HOLZ']['SOLZ'] = SOLZ
    tags['HOLZ']['HHOLZ'] = HOLZp  # even higher HOLZ

    tags['HOLZ']['hkl'] = tags['allowed']['hkl']
    tags['HOLZ']['intensities'] = intensities

    if verbose:
        print('KinsCat\'s  \"Kinematic_Scattering\" finished')


def plotSAED(tags, gray=False):
    """
    Plot SAED Pattern of single crystal
    """

    saed = tags.copy()
    saed['convergence_angle_nm-1'] = 0

    saed['background'] = 'white'   # 'white'  'grey'
    saed['color map'] = 'plasma'  # ,'cubehelix'#'Greys'#'plasma'
    saed['color reflections'] = 'ZOLZ'

    if gray:
        saed['color map'] = 'gray'
        saed['background'] = '#303030'  # 'darkgray'
        saed['color reflections'] = 'intensity'
    saed['plot HOLZ'] = 0
    saed['plot HOLZ excess'] = 0
    saed['plot Kikuchi'] = 0
    saed['plot reflections'] = 1

    saed['label HOLZ'] = 0
    saed['label Kikuchi'] = 0
    saed['label reflections'] = 0

    saed['label color'] = 'white'
    saed['label size'] = 10

    saed['color Laue Zones'] = ['red',  'blue', 'green', 'blue', 'green']  # , 'green', 'red'] #for OLZ give a sequence
    saed['color zero'] = 'red'  # 'None' #'white'
    saed['color ring zero'] = 'None'  # 'Red' #'white' #, 'None'
    saed['width ring zero'] = 2

    plot_diffraction_pattern(saed, True)


def plotKikuchi(tags, grey=False):
    """
    Plot Kikuchi Pattern
    """
    Kikuchi = tags.copy()

    Kikuchi['background'] = 'black'   # 'white'  'grey'
    Kikuchi['color map'] = 'plasma'   # ,'cubehelix'#'Greys'#'plasma'
    Kikuchi['color reflections'] = 'intensity'

    Kikuchi['plot HOLZ'] = 0
    Kikuchi['plot HOLZ excess'] = 0
    Kikuchi['plot Kikuchi'] = 1
    Kikuchi['plot reflections'] = 1

    Kikuchi['label HOLZ'] = 0
    Kikuchi['label Kikuchi'] = 0
    Kikuchi['label reflections'] = 0

    Kikuchi['label color'] = 'white'
    Kikuchi['label size'] = 10

    Kikuchi['color Kikuchi'] = 'green'
    Kikuchi['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    Kikuchi['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2
    Kikuchi['color Laue Zones'] = ['red',  'blue', 'green', 'blue', 'green']  # , 'green', 'red']
    # #for OLZ give a sequence
    Kikuchi['color zero'] = 'white'  # 'None' #'white'
    Kikuchi['color ring zero'] = 'None'  # 'Red' #'white' #, 'None'
    Kikuchi['width ring zero'] = 2

    plot_diffraction_pattern(Kikuchi, True)


def plotHOLZ(tags, grey=False):
    """
    Plot HOLZ Pattern
    """
    holz = tags.copy()

    holz['background'] = 'black'   # 'white'  'grey'
    holz['color map'] = 'plasma'   # ,'cubehelix'#'Greys'#'plasma'
    holz['color reflections'] = 'intensity'

    holz['plot HOLZ'] = 1
    holz['plot HOLZ excess'] = 0
    holz['plot Kikuchi'] = 1
    holz['plot reflections'] = 0

    holz['label HOLZ'] = 0
    holz['label Kikuchi'] = 0
    holz['label reflections'] = 0

    holz['label color'] = 'white'
    holz['label size'] = 10

    holz['color Kikuchi'] = 'green'
    holz['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    holz['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    holz['color Laue Zones'] = ['red',  'blue', 'green']  # , 'green', 'red'] #for OLZ give a sequence
    holz['color zero'] = 'white'   # 'None' #'white'
    holz['color ring zero'] = 'Red'  # 'Red' #'white' #, 'None'
    holz['width ring zero'] = 2

    plot_diffraction_pattern(holz, True)


def plotCBED(tags, grey=False):
    """
    Plot CBED Pattern
    """
    cbed = tags.copy()

    cbed['background'] = 'black'   # 'white'  'grey'
    cbed['color map'] = 'plasma'   # ,'cubehelix'#'Greys'#'plasma'
    cbed['color reflections'] = 'intensity'

    cbed['plot HOLZ'] = 1
    cbed['plot HOLZ excess'] = 1
    cbed['plot Kikuchi'] = 1
    cbed['plot reflections'] = 1

    cbed['label HOLZ'] = 0
    cbed['label Kikuchi'] = 0
    cbed['label reflections'] = 0

    cbed['label color'] = 'white'
    cbed['label size'] = 10

    cbed['color Kikuchi'] = 'green'
    cbed['linewidth HOLZ'] = -1  # -1: linewidth according to intensity (structure factor F^2
    cbed['linewidth Kikuchi'] = -1  # -1: linewidth according to intensity (structure factor F^2

    cbed['color reflections'] = 'intensity'

    cbed['color Laue Zones'] = ['red',  'blue', 'green']   # , 'green', 'red'] #for OLZ give a sequence
    cbed['color zero'] = 'white'  # 'None' #'white'
    cbed['color ring zero'] = 'Red'  # 'Red' #'white' #, 'None'
    cbed['width ring zero'] = 2

    plot_diffraction_pattern(cbed, True)


# #######################
# Plot HOLZ Pattern #
# #######################
def circles(x, y, s, c='b', vmin=None, vmax=None, **kwargs):
    """
    Make a scatter plot of circles.

    Similar to plt.scatter, but the size of circles are in data scale.

    Parameters
    ----------
    x, y : scalar or array_like, shape (n, )
        Input data
    s : scalar or array_like, shape (n, )
        Radius of circles.
    c : color or sequence of color, optional, default : 'b'
        `c` can be a single color format string, or a sequence of color
        specifications of length `N`, or a sequence of `N` numbers to be
        mapped to colors using the `cmap` and `norm` specified via kwargs.
        Note that `c` should not be a single numeric RGB or RGBA sequence
        because that is indistinguishable from an array of values
        to be colormapped. (If you insist, use `color` instead.)
        `c` can be a 2-D array in which the rows are RGB or RGBA, however.
    vmin, vmax : scalar, optional, default: None
        `vmin` and `vmax` are used in conjunction with `norm` to normalize
        luminance data.  If either are `None`, the min and max of the
        color array is used.
    kwargs : `~matplotlib.collections.Collection` properties
        Eg. alpha, edgecolor(ec), facecolor(fc), linewidth(lw), linestyle(ls),
        norm, cmap, transform, etc.

    Returns
    -------
    paths : ~matplotlib.collections.PathCollection

    Examples
    --------
    a = np.arange(11)
    circles(a, a, s=a*0.2, c=a, alpha=0.5, ec='none')
    plt.colorbar()

    License
    -------
    This code is under [The BSD 3-Clause License] (http://opensource.org/licenses/BSD-3-Clause)
    """

    if np.isscalar(c):
        kwargs.setdefault('color', c)
        c = None

    if 'fc' in kwargs:
        kwargs.setdefault('facecolor', kwargs.pop('fc'))
    if 'ec' in kwargs:
        kwargs.setdefault('edgecolor', kwargs.pop('ec'))
    if 'ls' in kwargs:
        kwargs.setdefault('linestyle', kwargs.pop('ls'))
    if 'lw' in kwargs:
        kwargs.setdefault('linewidth', kwargs.pop('lw'))
    # You can set `facecolor` with an array for each patch,
    # while you can only set `facecolors` with a value for all.

    zipped = np.broadcast(x, y, s)
    patches = [Circle((x_, y_), s_)
               for x_, y_, s_ in zipped]
    collection = PatchCollection(patches, **kwargs)
    if c is not None:
        c = np.broadcast_to(c, zipped.shape).ravel()
        collection.set_array(c)
        collection.set_clim(vmin, vmax)

    ax = plt.gca()
    ax.add_collection(collection)
    ax.autoscale_view()
    plt.draw_if_interactive()
    if c is not None:
        plt.sci(collection)
    return collection


def plot_diffraction_pattern(tags, grey=False):
    """
    Plot any diffraction pattern based on content in dictionary
    """
    if 'plot_rotation' not in tags:
        tags['plot_rotation'] = 0.

    if 'new_plot' not in tags:
        tags['new_plot'] = True

    if tags['new_plot']:
        fig = plt.figure()
        ax = fig.add_subplot(111, facecolor=tags['background'])

        diffraction_pattern(tags, grey)

        plt.axis('equal')
        if 'plot FOV' in tags:
            x = tags['plot FOV']  # in 1/nm
            plt.xlim(-x, x)
            plt.ylim(-x, x)
        plt.title(tags['crystal_name'])
        plt.show()
    else:
        diffraction_pattern(tags, grey)


def diffraction_pattern(tags, grey=False):
    """
    Determines how to plot diffraction pattern from kinematic scattering data.

    Parameters:
    -----------
    tags: dict
        dictionary of kinematic scattering data
    grey: boolean optional
        plot in gray scale

    Returns:
    --------
    tags: dict
        dictionary that now contains all information of how to plot any diffraction pattern
    plot: matplotlib figure
    """

    # Get information from dictionary
    HOLZ = tags['allowed']['HOLZ']
    ZOLZ = tags['allowed']['ZOLZ']

    Laue_Zone = tags['allowed']['Laue_Zone']

    if 'label' in tags['allowed']:
        label = tags['allowed']['label']

    angle = np.radians(tags['plot_rotation'])  # mrad
    c = np.cos(angle)
    s = np.sin(angle)
    r_mat = np.array([[c, -s, 0], [s, c, 0], [0, 0, 1]])

    # HOLZ and Kikuchi lines coordinates in Hough space
    gd = np.dot(tags['HOLZ']['g deficient'], r_mat)
    ge = np.dot(tags['HOLZ']['g excess'], r_mat)

    points = np.dot(tags['allowed']['g'], r_mat)

    theta = tags['HOLZ']['theta']+angle

    intensity = tags['allowed']['intensities']
    radius = tags['convergence_angle_nm-1']
    tags['Bragg'] = {}
    tags['Bragg']['points'] = points
    tags['Bragg']['intensity'] = intensity
    tags['Bragg']['radius'] = radius

    if radius < 0.1:
        radiusI = 2
    else:
        radiusI = radius
    # Beginning and ends of HOLZ lines
    maxlength = radiusI * 1.3
    Hxp = gd[:, 0] + maxlength * np.cos(np.pi-theta)
    Hyp = gd[:, 1] + maxlength * np.sin(np.pi-theta)
    Hxm = gd[:, 0] - maxlength * np.cos(np.pi-theta)
    Hym = gd[:, 1] - maxlength * np.sin(np.pi-theta)
    tags['HOLZ lines'] = {}
    tags['HOLZ lines']['Hxp'] = Hxp
    tags['HOLZ lines']['Hyp'] = Hyp
    tags['HOLZ lines']['Hxm'] = Hxm
    tags['HOLZ lines']['Hym'] = Hym

    # Beginning and ends of excess HOLZ lines
    maxlength = radiusI * 0.8
    Exp = ge[:, 0] + maxlength * np.cos(np.pi-theta)
    Eyp = ge[:, 1] + maxlength * np.sin(np.pi-theta)
    Exm = ge[:, 0] - maxlength * np.cos(np.pi-theta)
    Eym = ge[:, 1] - maxlength * np.sin(np.pi-theta)
    tags['excess HOLZ lines'] = {}
    tags['excess HOLZ lines']['Exp'] = Exp
    tags['excess HOLZ lines']['Eyp'] = Eyp
    tags['excess HOLZ lines']['Exm'] = Exm
    tags['excess HOLZ lines']['Eym'] = Eym

    # Beginning and ends of HOLZ lines
    maxlength = 20
    Kxp = gd[:, 0] + maxlength * np.cos(np.pi-theta)
    Kyp = gd[:, 1] + maxlength * np.sin(np.pi-theta)
    Kxm = gd[:, 0] - maxlength * np.cos(np.pi-theta)
    Kym = gd[:, 1] - maxlength * np.sin(np.pi-theta)
    tags['Kikuchi lines'] = {}
    tags['Kikuchi lines']['Kxp'] = Kxp
    tags['Kikuchi lines']['Kyp'] = Kyp
    tags['Kikuchi lines']['Kxm'] = Kxm
    tags['Kikuchi lines']['Kym'] = Kym

    intensity_Kikuchi = intensity*4./intensity[ZOLZ].max()
    if len(intensity[tags['HOLZ']['HOLZ']]) > 1:
        intensity_HOLZ = intensity*4./intensity[tags['HOLZ']['HOLZ']].max()*.75
        tags['HOLZ lines']['intensity_HOLZ'] = intensity_HOLZ
    tags['Kikuchi lines']['intensity_Kikuchi'] = intensity_Kikuchi
    # #######
    # Plot  #
    # #######
    cms = mpl.cm
    # cm = cms.plasma#jet#, cms.gray, cms.autumn]
    cm = plt.get_cmap(tags['color map'])

    fig = plt.gcf()
    ax = plt.gca()

    if 'plot image' in tags:
        l = -tags['plot image FOV']/2+tags['plot shift x']
        r = tags['plot image FOV']/2+tags['plot shift x']
        t = -tags['plot image FOV']/2+tags['plot shift y']
        b = tags['plot image FOV']/2+tags['plot shift y']
        ax.imshow(tags['plot image'], extent=(l, r, t, b))
        print('image')

    ix = np.argsort((points**2).sum(axis=1))
    # print(tags['allowed']['hkl'][ix])
    p = points[ix]
    inten = intensity[ix]
    tags['Bragg']['points'] = p
    tags['Bragg']['intensity'] = inten

    Lauecolor = []
    for i in range(int(Laue_Zone.max())+1):
        if i < len(tags['color Laue Zones']):
            Lauecolor.append(tags['color Laue Zones'][i])
        else:
            Lauecolor.append(tags['color Laue Zones'][-1])

    if tags['plot reflections']:
        if radius < 0.1:
            if tags['color reflections'] == 'intensity':
                ax.scatter(points[:, 0], points[:, 1], c=np.log(intensity), cmap=cm, s=20)
            else:
                for i in range(len(Laue_Zone)):
                    color = Lauecolor[int(Laue_Zone[i])]
                    ax.scatter(points[i, 0], points[i, 1], c=color, cmap=cm, s=20)

            ax.scatter(0, 0, c=tags['color zero'], s=100)
            radius = 2
        else:
            ix = np.argsort((points**2).sum(axis=1))
            p = points[ix]
            inten = intensity[ix]
            if tags['color reflections'] == 'intensity':
                circles(p[:, 0], p[:, 1], s=radius, c=np.log(inten+1), cmap=cm, alpha=0.9, edgecolor='')
            else:
                for i in range(len(Laue_Zone)):
                    color = Lauecolor[int(Laue_Zone[i])]
                    circles(p[i, 0], p[i, 1], s=radius, c=color, cmap=cm, alpha=0.9, edgecolor='')

    if not tags['color zero'] == 'None':
        circle = plt.Circle((0, 0), radius, color=tags['color zero'])
        ax.add_artist(circle)

    for i in range(len(Hxp)):
        if tags['HOLZ']['HOLZ'][i]:
            color = Lauecolor[int(Laue_Zone[i])]
            if tags['plot HOLZ']:
                # plot HOLZ lines
                ax.plot((Hxp[i], Hxm[i]), (Hyp[i], Hym[i]), c=color, linewidth=intensity_HOLZ[i])
            if tags['plot HOLZ excess']:
                ax.plot((Exp[i], Exm[i]), (Eyp[i], Eym[i]), c=color, linewidth=intensity_HOLZ[i])

                if tags['label HOLZ']:  # Add indices
                    ax.text(Hxp[i], Hyp[i], label[i], fontsize=10)
                    ax.text(Exp[i], Eyp[i], label[i], fontsize=10)
        else:
            # Plot Kikuchi lines
            if tags['plot Kikuchi']:
                ax.plot((Kxp[i], Kxm[i]), (Kyp[i], Kym[i]), c=tags['color Kikuchi'], linewidth=intensity_Kikuchi[i])
                if tags['label Kikuchi']:  # Add indices
                    ax.text(Kxp[i], Kyp[i], label[i], fontsize=tags['label size'], color=tags['label color'])
                tags['Kikuchi lines']

    if not (tags['color ring zero'] == 'None'):
        ring = plt.Circle((0, 0), radius, color=tags['color ring zero'], fill=False, linewidth=2)
        ax.add_artist(ring)
        print(ring)
    return tags


def feq(element, q):
    """Atomic form factor parametrized in 1/Angstrom but converted to 1/nm

    The atomic form factor is from Kirkland: Advanced Computing in Electron Microscopy 2nd edition, Appendix C.
    From Appendix C of Kirkland, "Advanced Computing in Electron Microscopy", 2nd ed.
    Calculation of electron form factor for specific q:
    Using equation Kirkland C.15

    Parameters
    ----------
    element: string
        element name
    q: float
        magnitude of scattering vector in 1/nm -- (=> exp(-i*g.r), physics negative convention)

    Returns
    -------
    fL+fG: float
        atomic scattering vector
    """

    if not isinstance(element, str):
        raise TypeError('Element has to be a string')
    if element not in electronFF:
        if len(element) > 2:
            raise TypeError('Please use standard convention for element abbreviation with not more than two letters')
        else:
            raise TypeError('Element {element} not known to electron diffraction should')
    if not isinstance(q, (float, int)):
        raise TypeError('Magnitude of scattering vector has to be a number of type float')

    q = q/10
    # q is now magnitude of scattering vector in 1/A -- (=> exp(-i*g.r), physics negative convention)
    param = electronFF[element]
    fL = 0
    fG = 0
    for i in range(3):
        fL += param['fa'][i]/(q**2 + param['fb'][i])
        fG += param['fc'][i]*np.exp(-q**2 * param['fd'][i])

    # Conversion factor from scattering factors to volts. h^2/(2pi*m0*e), see e.g. Kirkland eqn. C.5
    # !NB RVolume is already in A unlike RPlanckConstant
    # ScattFacToVolts=(PlanckConstant**2)*(AngstromConversion**2)/(2*np.pi*ElectronMass*ElectronCharge)
    return fL+fG  # * ScattFacToVolts



"""Will move to sidpy"""
from PyQt5 import QtWidgets, QtCore


class ProgressDialog(QtWidgets.QDialog):
    """Simple dialog that consists of a Progress Bar and a Button.

    Clicking on the button results in the start of a timer and
    updates the progress bar.
    """

    def __init__(self, title='Progress', number_of_counts=100):
        super().__init__()
        self.setWindowTitle(title)
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)

        self.progress = QtWidgets.QProgressBar(self)
        self.progress.setGeometry(10, 10, 500, 50)
        self.progress.setMaximum(number_of_counts)
        self.show()

    def set_value(self, count):
        self.progress.setValue(count)
        QtWidgets.QApplication.processEvents()



"""
GUI definitions for EEELS_dialog
"""

# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'untitled.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class UiDialog(object):
    """ GUI definitions"""
    def __init__(self, dialog):
        dialog.setObjectName('Fit Composition Input')
        dialog.resize(371, 184)

        valid_float = QtGui.QDoubleValidator()
        valid_int = QtGui.QIntValidator()
        valid_int.setBottom(0)

        self.layout = QtWidgets.QGridLayout()
        self.layout.setVerticalSpacing(2)

        self.separator1 = QtWidgets.QLabel(dialog)
        self.separator1.setAutoFillBackground(True)
        palette = self.separator1.palette()
        palette.setColor(self.separator1.backgroundRole(), QtCore.Qt.blue)
        palette.setColor(self.separator1.foregroundRole(), QtCore.Qt.white)
        self.separator1.setAlignment(QtCore.Qt.AlignCenter)
        self.separator1.setMaximumHeight(50)

        self.separator1.setPalette(palette)
        row = 0
        ######################################################################
        self.separator1.setText("Fit Region")
        self.layout.addWidget(self.separator1, row, 0, 1, 4)
        ######################################################################

        name = 'fit_start'

        self.label1 = QtWidgets.QLabel('Fit Start:')
        self.edit1 = QtWidgets.QLineEdit('0')
        self.edit1.setValidator(valid_float)
        self.unit1 = QtWidgets.QLabel('eV')
        self.edit1.setObjectName(name+'_edit')
        self.unit1.setObjectName(name+'_units')

        row += 1
        self.layout.addWidget(self.label1, row, 0)
        self.layout.addWidget(self.edit1, row, 1)
        self.layout.addWidget(self.unit1, row, 2)

        name = 'fit_end'

        self.label2 = QtWidgets.QLabel('Fit End:')
        self.edit2 = QtWidgets.QLineEdit('0')
        self.edit2.setValidator(valid_float)
        self.unit2 = QtWidgets.QLabel('eV')
        self.edit2.setObjectName(name + '_edit')
        self.unit2.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label2, row, 0)
        self.layout.addWidget(self.edit2, row, 1)
        self.layout.addWidget(self.unit2, row, 2)

        self.separator2 = QtWidgets.QLabel(dialog)
        self.separator2.setAutoFillBackground(True)
        self.separator2.setAlignment(QtCore.Qt.AlignCenter)
        self.separator2.setMaximumHeight(50)
        self.separator2.setPalette(palette)

        row += 1
        ######################################################################
        self.separator2.setText("Elements")
        self.layout.addWidget(self.separator2, row, 0, 1, 4)
        ######################################################################

        name = 'edge'
        self.label3 = QtWidgets.QLabel('Edge:')
        self.edge_list = ['Edge 1', 'Edge 2', 'Edge 3', 'Edge 4', 'Edge 5', 'Edge 6']
        self.list3 = QtWidgets.QComboBox()
        self.list3.setEditable(False)
        self.list3.addItems(self.edge_list)
        self.check3 = QtWidgets.QCheckBox("Regions")

        row += 1
        self.layout.addWidget(self.label3, row, 0)
        self.layout.addWidget(self.list3, row, 1)
        self.layout.addWidget(self.check3, row, 2)

        self.list3.setObjectName(name + '_list')
        self.check3.setObjectName(name + '_check')

        name = 'element'

        self.label4 = QtWidgets.QLabel('Z:')
        self.edit4 = QtWidgets.QLineEdit('1')
        self.unit4 = QtWidgets.QLabel('H')
        self.edit4.setObjectName(name + '_edit')
        self.unit4.setObjectName(name + '_name')

        row += 1
        self.layout.addWidget(self.label4, row, 0)
        self.layout.addWidget(self.edit4, row, 1)
        self.layout.addWidget(self.unit4, row, 2)

        name = 'symmetry'
        self.edge_sym = ['K1', 'L3', 'M5', 'M3', 'M1', 'N1', 'N3', 'N5', 'N7', 'O1', 'O3', 'O5']

        self.label5 = QtWidgets.QLabel('Symmetry:')
        self.list5 = QtWidgets.QComboBox()
        self.list5.setEditable(False)
        self.list5.addItems(self.edge_sym)

        self.x_sec_meth = ['XRPA', 'Hydrog.', 'H-S GOS']
        self.select5 = QtWidgets.QComboBox()
        self.select5.setEditable(False)
        self.select5.addItems(self.x_sec_meth)

        row += 1
        self.layout.addWidget(self.label5, row, 0)
        self.layout.addWidget(self.list5, row, 1)
        self.layout.addWidget(self.select5, row, 2)

        self.list5.setObjectName(name + '_list')
        self.select5.setObjectName(name + '_method')

        name = 'onset'

        self.label6 = QtWidgets.QLabel('Onset:')
        self.edit6 = QtWidgets.QLineEdit('0')
        self.edit6.setValidator(valid_float)
        self.unit6 = QtWidgets.QLabel('eV')
        self.edit6.setObjectName(name + '_edit')
        self.unit6.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label6, row, 0)
        self.layout.addWidget(self.edit6, row, 1)
        self.layout.addWidget(self.unit6, row, 2)

        name = 'excl_start'

        self.label7 = QtWidgets.QLabel('Excl. Start:')
        self.edit7 = QtWidgets.QLineEdit('0')
        self.edit7.setValidator(valid_float)
        self.unit7 = QtWidgets.QLabel('eV')
        self.edit7.setObjectName(name + '_edit')
        self.unit7.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label7, row, 0)
        self.layout.addWidget(self.edit7, row, 1)
        self.layout.addWidget(self.unit7, row, 2)

        name = 'excl_end'

        self.label8 = QtWidgets.QLabel('Excl. End:')
        self.edit8 = QtWidgets.QLineEdit('0')
        self.edit8.setValidator(valid_float)
        self.unit8 = QtWidgets.QLabel('eV')
        self.edit8.setObjectName(name + '_edit')
        self.unit8.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label8, row, 0)
        self.layout.addWidget(self.edit8, row, 1)
        self.layout.addWidget(self.unit8, row, 2)

        name = 'multiplier'

        self.label9 = QtWidgets.QLabel('Multiplier:')
        self.edit9 = QtWidgets.QLineEdit('0')
        self.edit9.setValidator(valid_float)
        self.unit9 = QtWidgets.QLabel('a.u.')
        self.edit9.setObjectName(name + '_edit')
        self.unit9.setObjectName(name + '_units')

        row += 1
        self.layout.addWidget(self.label9, row, 0)
        self.layout.addWidget(self.edit9, row, 1)
        self.layout.addWidget(self.unit9, row, 2)

        self.separator3 = QtWidgets.QLabel(dialog)
        self.separator3.setAutoFillBackground(True)
        self.separator3.setAlignment(QtCore.Qt.AlignCenter)
        self.separator3.setMaximumHeight(50)
        self.separator3.setPalette(palette)
        ######################################################################
        self.separator3.setText("Other Fit Parameters")

        row += 1
        self.layout.addWidget(self.separator3, row, 0, 1, 4)
        ######################################################################

        self.check10 = QtWidgets.QCheckBox("Conv. LL")
        self.show_edges = QtWidgets.QCheckBox("Show Edges")
        self.select10 = QtWidgets.QCheckBox("Fit Comp.")

        row += 1
        self.layout.addWidget(self.check10, row, 0)
        self.layout.addWidget(self.show_edges, row, 1)
        self.layout.addWidget(self.select10, row, 2)

        self.check10.setObjectName('conv_ll')
        self.select10.setObjectName('fit_comp')

        row += 1
        self.progress = QtWidgets.QProgressBar()
        self.progress.setGeometry(0, 0, 300, 25)
        self.progress.setMaximum(100)

        self.do_all_button = QtWidgets.QPushButton('Do All', dialog)

        self.do_fit_button = QtWidgets.QPushButton('Fit Composition', dialog)
        self.do_fit_button.setCheckable(True)

        self.layout.addWidget(self.do_all_button, row, 0)
        self.layout.addWidget(self.progress, row, 1)
        self.layout.addWidget(self.do_fit_button, row, 2)

        dialog.setLayout(self.layout)
        dialog.setWindowTitle('Fit Composition Input')

        # self.buttonBox.accepted.connect(Dialog.accept)
        # self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(dialog)



"""plotting of sidpy Datasets with bokeh for google colab"""

import numpy as np
import sidpy

from bokeh.layouts import column
from bokeh.plotting import figure  # , show, output_notebook
from bokeh.models import CustomJS, Slider, Span
from bokeh.models import LinearColorMapper, ColorBar, ColumnDataSource, BoxSelectTool
from bokeh.palettes import Spectral11

from pyTEMlib.sidpy_tools import *
import sys
import matplotlib.pyplot as plt
# from matplotlib.widgets import Slider, Button
import matplotlib.patches as patches
# import matplotlib.animation as animation

if sys.version_info.major == 3:
    unicode = str

default_cmap = plt.cm.viridis


def plot(dataset, palette='Viridis256'):
    """plot according to data_type"""
    if dataset.data_type.name == 'IMAGE_STACK':
        p = plot_stack(dataset, palette=palette)
    elif dataset.data_type.name == 'IMAGE':
        p = plot_image(dataset, palette=palette)
    elif dataset.data_type.name == 'SPECTRUM':
        p = plot_spectrum(dataset, palette=palette)
    else:
        p = None
    return p


def plot_stack(dataset, palette="Viridis256"):
    """Plotting a stack of images

    Plotting a stack of images contained in a sidpy.Dataset.
    The images can be scrolled through with a slider widget.

    Parameters
    ----------
    dataset: sidpy.Dataset
        sidpy dataset with data_type 'IMAGE_STACK'
    palette: bokeh palette
        palette is optional

    Returns
    -------
    p: bokeh plot

    Example
    -------
    >> import pyTEMlib
    >> from bokeh.plotting import figure, show, output_notebook
    >> output_notebook()
    >> p = pyTEMlib.viz(dataset)
    >> p.show(p)
    """

    if not isinstance(dataset, sidpy.Dataset):
        raise TypeError('Need a sidpy dataset for plotting')
    if dataset.data_type.name != 'IMAGE_STACK':
        raise TypeError('Need an IMAGE_STACK for plotting a stack')

    stack = np.array(dataset-dataset.min())
    stack = stack/stack.max()*256
    stack = np.array(stack, dtype=int)

    color_mapper = LinearColorMapper(palette=palette, low=0, high=256)

    p = figure(match_aspect=True, plot_width=600, plot_height=600)
    im_plot = p.image(image=[stack[0]], x=[0], y=[0], dw=[dataset.x[-1]], dh=[dataset.y[-1]], color_mapper=color_mapper)
    p.x_range.range_padding = 0
    p.y_range.range_padding = 0
    p.xaxis.axis_label = 'distance (nm)'
    p.yaxis.axis_label = 'distance (nm)'

    slider = Slider(start=0, end=stack.shape[0]-1, value=0, step=1, title="frame")

    update_curve = CustomJS(args=dict(source=im_plot, slider=slider, stack=stack),
                            code="""var f = slider.value;
                                    source.data_source.data['image'] = [stack[f]];
                                    // necessary because we mutated source.data in-place
                                    source.data_source.change.emit(); """)
    slider.js_on_change('value', update_curve)

    return column(slider, p)


def plot_image(dataset, palette="Viridis256"):
    """Plotting an image

    Plotting an image contained in a sidpy.Dataset.

    Parameters
    ----------
    dataset: sidpy.Dataset
        sidpy dataset with data_type 'IMAGE_STACK'
    palette: bokeh palette
        palette is optional

    Returns
    -------
    p: bokeh plot

    Example
    -------
    >> import pyTEMlib
    >> from bokeh.plotting import figure, show, output_notebook
    >> output_notebook()
    >> p = pyTEMlib.viz(dataset)
    >> p.show(p)


        """
    if not isinstance(dataset, sidpy.Dataset):
        raise TypeError('Need a sidpy dataset for plotting')

    if dataset.data_type.name not in ['IMAGE', 'IMAGE_STACK']:
        raise TypeError('Need an IMAGE or IMAGE_STACK for plotting an image')

    if dataset.data_type.name == 'IMAGE_STACK':
        image = dataset.sum(axis=0)
        image = sidpy.Dataset.from_array(image)
        image.data_type = 'image'
        image.title = dataset.title
        image.set_dimension(0, dataset.dim_1)
        image.set_dimension(1, dataset.dim_2)
    else:
        image = dataset

    p = figure(tooltips=[("x", "$x"), ("y", "$y"), ("value", "@image")], match_aspect=True,
               plot_width=675, plot_height=600, )
    color_mapper = LinearColorMapper(palette=palette, low=float(image.min()), high=float(image.max()))

    # must give a vector of image data for image parameter
    p.image(image=[np.array(image)], x=0, y=0, dw=image.x[-1], dh=image.y[-1], color_mapper=color_mapper,
            level="image")
    p.x_range.range_padding = 0
    p.y_range.range_padding = 0

    p.grid.grid_line_width = 0.
    p.xaxis.axis_label = 'distance (nm)'
    p.yaxis.axis_label = 'distance (nm)'

    color_bar = ColorBar(color_mapper=color_mapper, major_label_text_font_size="7pt",
                         label_standoff=6, border_line_color=None, location=(0, 0))
    p.add_layout(color_bar, 'right')
    return p


def plot_spectrum(dataset, selected_range, palette=Spectral11):
    """Plot spectrum"""
    if not isinstance(dataset, sidpy.Dataset):
        raise TypeError('Need a sidpy dataset for plotting')

    if dataset.data_type.name not in ['SPECTRUM']:
        raise TypeError('Need an sidpy.Dataset of data_type SPECTRUM for plotting a spectrum ')

    p = figure(x_axis_type="linear", plot_width=800, plot_height=400,
               tooltips=[("index", "$index"), ("(x,y)", "($x, $y)")],
               tools="pan,wheel_zoom,box_zoom,reset, hover, lasso_select")
    p.add_tools(BoxSelectTool(dimensions="width"))

    # first line is dataset
    spectrum = ColumnDataSource(data=dict(x=dataset.dim_0, y=np.array(dataset)))
    p.scatter('x', 'y', color='blue', size=1, alpha=0., source=spectrum,
              selection_color="firebrick", selection_alpha=0.)
    p.line(x='x', y='y', source=spectrum, legend_label=dataset.title, color=palette[0], line_width=2)
    # add other lines if available
    if 'add2plot' in dataset.metadata:
        data = dataset.metadata['add2plot']
        for key, line in data.items():
            p.line(dataset.dim_0.values, line['data'], legend_label=line['legend'], color=palette[key], line_width=2)
    p.legend.click_policy = "hide"
    p.xaxis.axis_label = dataset.labels[0]
    p.yaxis.axis_label = dataset.data_descriptor
    p.title.text = dataset.title

    my_span = Span(location=0, dimension='width', line_color='gray', line_width=1)
    p.add_layout(my_span)

    callback = CustomJS(args=dict(s1=spectrum), code="""
        var inds = s1.selected.indices;
        if (inds.length == 0)
            return;
        var kernel = IPython.notebook.kernel;
        kernel.execute("selected_range = " + [inds[0], inds[inds.length-1]]);""")

    spectrum.selected.js_on_change('indices', callback)
    return p


class CurveVisualizer(object):
    """Plots a sidpy.Dataset with spectral dimension

    """
    def __init__(self, dset, spectrum_number=None, axis=None, leg=None, **kwargs):
        if not isinstance(dset, sidpy.Dataset):
            raise TypeError('dset should be a sidpy.Dataset object')
        if axis is None:
            self.fig = plt.figure()
            self.axis = self.fig.add_subplot(1, 1, 1)
        else:
            self.axis = axis
            self.fig = axis.figure

        self.dset = dset
        self.selection = []
        [self.spec_dim, self.energy_scale] = get_dimensions_by_type('spectral', self.dset)[0]

        self.lined = dict()
        self.plot(**kwargs)

    def plot(self, **kwargs):
        line1, = self.axis.plot(self.energy_scale.values, self.dset, label='spectrum', **kwargs)
        lines = [line1]
        if 'add2plot' in self.dset.metadata:
            data = self.dset.metadata['add2plot']
            for key, line in data.items():
                line_add, = self.axis.plot(self.energy_scale.values,  line['data'], label=line['legend'])
                lines.append(line_add)

            legend = self.axis.legend(loc='upper right', fancybox=True, shadow=True)
            legend.get_frame().set_alpha(0.4)

            for legline, origline in zip(legend.get_lines(), lines):
                legline.set_picker(True)
                legline.set_pickradius(5)  # 5 pts tolerance
                self.lined[legline] = origline
            self.fig.canvas.mpl_connect('pick_event', self.onpick)

        self.axis.axhline(0, color='gray', alpha=0.6)
        self.axis.set_xlabel(self.dset.labels[0])
        self.axis.set_ylabel(self.dset.data_descriptor)
        self.axis.ticklabel_format(style='sci', scilimits=(-2, 3))
        self.fig.canvas.draw_idle()

    def update(self, **kwargs):
        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.axis.clear()
        self.plot(**kwargs)
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)

    def onpick(self, event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        origline = self.lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legline.set_alpha(1.0)
        else:
            legline.set_alpha(0.2)
        self.fig.canvas.draw()



"""
Input Dialog for EELS Analysis

Author: Gerd Duscher

"""
import numpy as np
import sidpy

from PyQt5 import QtCore,  QtWidgets

from pyTEMlib import info_dlg
import pyTEMlib.interactive_eels as ieels
from pyTEMlib.microscope import microscope

from pyTEMlib import file_tools as ft
_version = 000


class InfoDialog(QtWidgets.QDialog):
    """
    Input Dialog for EELS Analysis

    Opens a PyQt5 GUi Dialog that allows to set the experimental parameter necessary for a Quantification.


    The dialog operates on a sidpy dataset
    """

    def __init__(self, dataset=None):
        super().__init__(None, QtCore.Qt.WindowStaysOnTopHint)
        # Create an instance of the GUI
        self.ui = info_dlg.UiDialog(self)
        self.set_action()

        self.dataset = dataset
        self.spec_dim = []
        self.energy_scale = np.array([])
        self.experiment = {}
        self.energy_dlg = None
        self.axis = None
        self.show()

        if dataset is None:
            # make a dummy dataset for testing
            dataset = ft.make_dummy_dataset(sidpy.DataType.SPECTRUM)
        if not isinstance(dataset, sidpy.Dataset):
            raise TypeError('dataset has to be a sidpy dataset')

        self.set_dataset(dataset)

        self.dataset.plot()
        if hasattr(self.dataset.view, 'axes'):
            self.axis = self.dataset.view.axes[-1]
        elif hasattr(self.dataset.view, 'axis'):
            self.axis = self.dataset.view.axis
        self.figure = self.axis.figure
        self.plot()
        self.update()

    def set_dataset(self, dataset):
        self.dataset = dataset
        if not hasattr(self.dataset, '_axes'):
            self.dataset._axes = self.dataset.axes
        if not hasattr(self.dataset, 'meta_data'):
            self.dataset.meta_data = {}

        self.spec_dim = ft.get_dimensions_by_type(sidpy.DimensionType.SPECTRAL, dataset)
        if len(self.spec_dim) != 1:
            raise TypeError('We need exactly one SPECTRAL dimension')
        self.spec_dim = self.spec_dim[0]
        self.energy_scale = self.spec_dim[1].values.copy()

        minimum_info = {'offset': self.energy_scale[0],
                        'dispersion': self.energy_scale[1] - self.energy_scale[0],
                        'exposure_time': 0.0,
                        'convergence_angle': 0.0, 'collection_angle': 0.0,
                        'acceleration_voltage': 100.0, 'binning': 1, 'conversion': 1.0,
                        'flux': 1.0, 'current': 1.0, 'SI_bin_x': 1, 'SI_bin_y': 1}
        if 'experiment' not in self.dataset.metadata:
            self.dataset.metadata['experiment'] = minimum_info
        self.experiment = self.dataset.metadata['experiment']

        for key, item in minimum_info.items():
            if key not in self.experiment:
                self.experiment[key] = item

    def set_dimension(self):
        self.spec_dim = ft.get_dimensions_by_type(sidpy.DimensionType.SPECTRAL, self.dataset)
        self.spec_dim = self.spec_dim[0]
        old_energy_scale = self.spec_dim[1]
        self.dataset.set_dimension(self.spec_dim[0], sidpy.Dimension(np.array(self.energy_scale),
                                                                     name=old_energy_scale.name,
                                                                     dimension_type=sidpy.DimensionType.SPECTRAL,
                                                                     units='eV',
                                                                     quantity='energy loss'))

    def update(self):

        self.ui.offsetEdit.setText(f"{self.experiment['offset']:.3f}")
        self.ui.dispersionEdit.setText(f"{self.experiment['dispersion']:.3f}")
        self.ui.timeEdit.setText(f"{self.experiment['exposure_time']:.6f}")

        self.ui.convEdit.setText(f"{self.experiment['convergence_angle']:.2f}")
        self.ui.collEdit.setText(f"{self.experiment['collection_angle']:.2f}")
        self.ui.E0Edit.setText(f"{self.experiment['acceleration_voltage']/1000.:.2f}")

        self.ui.binningEdit.setText(f"{self.experiment['binning']}")
        self.ui.conversionEdit.setText(f"{self.experiment['conversion']:.2f}")
        self.ui.fluxEdit.setText(f"{self.experiment['flux']:.2f}")
        self.ui.VOAEdit.setText(f"{self.experiment['current']:.2f}")

    def on_enter(self):
        sender = self.sender()

        if sender == self.ui.offsetEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['offset'] = value
            sender.setText(f"{value:.2f}")
            self.energy_scale = self.energy_scale - self.energy_scale[0] + value
            self.set_dimension()
            self.plot()
        elif sender == self.ui.dispersionEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['dispersion'] = value
            self.energy_scale = np.arange(len(self.energy_scale)) * value + self.energy_scale[0]
            self.set_dimension()
            self.plot()
            sender.setText(f"{value:.3f}")
        elif sender == self.ui.timeEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['exposure_time'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.convEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['convergence_angle'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.collEdit:
            value = float(str(sender.displayText()).strip())
            self.experiment['collection_angle'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.E0Edit:
            value = float(str(sender.displayText()).strip())
            self.experiment['acceleration_voltage'] = value*1000.0
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.E0Edit:
            value = float(str(sender.displayText()).strip())
            self.experiment['acceleration_voltage'] = value
            sender.setText(f"{value:.2f}")
        elif sender == self.ui.binXEdit or sender == self.ui.binYEdit:
            if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
                bin_x = int(self.ui.binXEdit.displayText())
                bin_y = int(self.ui.binYEdit.displayText())
                self.experiment['SI_bin_x'] = bin_x
                self.experiment['SI_bin_y'] = bin_y
                self.dataset.view.set_bin([bin_x, bin_y])
                self.ui.binXEdit.setText(str(self.dataset.view.bin_x))
                self.ui.binYEdit.setText(str(self.dataset.view.bin_y))
        else:
            print('not supported yet')

    def plot(self):
        if self.dataset.data_type == sidpy.DataType.SPECTRAL_IMAGE:
            spectrum = self.dataset.view.get_spectrum()
            self.axis = self.dataset.view.axes[1]
        else:
            spectrum = np.array(self.dataset)
            self.axis = self.dataset.view.axis

        x_limit = self.axis.get_xlim()
        y_limit = self.axis.get_ylim()
        self.axis.clear()

        self.axis.plot(self.energy_scale, spectrum, label='spectrum')
        self.axis.set_xlim(x_limit)
        self.axis.set_ylim(y_limit)
        self.figure.canvas.draw_idle()

    def on_list_enter(self):
        sender = self.sender()
        if sender == self.ui.TEMList:
            microscope.set_microscope(self.ui.TEMList.currentText())
            self.experiment['microscope'] = microscope.name
            self.experiment['convergence_angle'] = microscope.alpha
            self.experiment['collection_angle'] = microscope.beta
            self.experiment['acceleration_voltage'] = microscope.E0
            self.update()

    def set_energy_scale(self):
        self.energy_dlg = ieels.EnergySelector(self.dataset)

        self.energy_dlg.signal_selected[bool].connect(self.set_energy)
        self.energy_dlg.show()

    def set_energy(self, k):
        self.spec_dim = ft.get_dimensions_by_type(sidpy.DimensionType.SPECTRAL, self.dataset)
        self.spec_dim = self.spec_dim[0]

        self.energy_scale = self.spec_dim[1]
        self.experiment['offset'] = self.energy_scale[0]
        self.experiment['dispersion'] = self.energy_scale[1] - self.energy_scale[0]
        self.update()

    def on_check(self):
        pass

    def set_action(self):
        self.ui.offsetEdit.editingFinished.connect(self.on_enter)
        self.ui.dispersionEdit.editingFinished.connect(self.on_enter)
        self.ui.timeEdit.editingFinished.connect(self.on_enter)

        self.ui.TEMList.activated[str].connect(self.on_list_enter)

        self.ui.convEdit.editingFinished.connect(self.on_enter)
        self.ui.collEdit.editingFinished.connect(self.on_enter)
        self.ui.E0Edit.editingFinished.connect(self.on_enter)
        self.ui.binningEdit.editingFinished.connect(self.on_enter)
        self.ui.conversionEdit.editingFinished.connect(self.on_enter)
        self.ui.fluxEdit.editingFinished.connect(self.on_enter)
        self.ui.VOAEdit.editingFinished.connect(self.on_enter)
        self.ui.energy_button.clicked.connect(self.set_energy_scale)
        self.ui.binXEdit.editingFinished.connect(self.on_enter)
        self.ui.binYEdit.editingFinished.connect(self.on_enter)



"""
Created on Sat Jan 19 09:08:25 2019

@author: gduscher
"""
# -*- coding: utf-8 -*-

from codecs import open
import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

with open(os.path.join(here, 'pyTEMlib/version.py')) as f:
    __version__ = f.read().split("'")[1]

setuptools.setup(
    name="pyTEMlib",
    version=__version__,
    description="pyTEM: TEM Data Quantification library through a model-based approach",
    long_description=long_description,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis'],
    keywords=['imaging', 'spectra', 'transmission', 'electron', 'microscopy',
              'scientific', 'scanning', 'eels', 'visualization', 'processing',
              'storage', 'hdf5'],
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*",
                                               "tests.*", "tests"]),
    url="https://web.utk.edu/~gduscher/Quantifit/",
    license='MIT',
    author="Gerd Duscher",
    author_email="gduscher@utk.edu",
    install_requires=['scipy', 'numpy', 'pillow', 'simpleITK', 'ase',
                      'scikit-image', 'scikit-learn', 'pyNSID', 'sidpy', 'SciFiReaders'],  # 'PyQt5> 1.0'],#
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    platforms=['Linux', 'Mac OSX', 'Windows 10/8.1/8/7'],
    package_data={"pyTEMlib": ["data/*"]},
    test_suite='pytest',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pyTEMlib=pyTEMlib:main',
        ],
    },
)



"""Functions to calculate electron probe"""
import numpy as np


def make_gauss(size_x, size_y, width=1.0, x0=0.0, y0=0.0, intensity=1.0):
    """Make a Gaussian shaped probe """
    size_x = size_x/2
    size_y = size_y/2
    x, y = np.mgrid[-size_x:size_x, -size_y:size_y]
    g = np.exp(-((x-x0)**2 + (y-y0)**2) / 2.0 / width**2)
    probe = g / g.sum() * intensity

    return probe


def make_lorentz(size_x, size_y, gamma=1.0, x0=0., y0=0., intensity=1.):
    """Make a Lorentzian shaped probe """

    size_x = np.floor(size_x / 2)
    size_y = np.floor(size_y / 2)
    x, y = np.mgrid[-size_x:size_x, -size_y:size_y]
    g = gamma / (2*np.pi) / np.power(((x-x0)**2 + (y-y0)**2 + gamma**2), 1.5)
    probe = g / g.sum() * intensity
    return probe


def zero_loss_peak_weight():
    # US100 zero_loss peak for Cc of aberrations
    x = np.linspace(-0.5, 0.9, 29)
    y = [0.0143, 0.0193, 0.0281, 0.0440, 0.0768, 0.1447, 0.2785, 0.4955, 0.7442, 0.9380, 1.0000, 0.9483, 0.8596,
         0.7620, 0.6539, 0.5515, 0.4478, 0.3500, 0.2683, 0.1979, 0.1410, 0.1021, 0.0752, 0.0545, 0.0401, 0.0300,
         0.0229, 0.0176, 0.0139]
    return x, y


# ## Aberration Function for Probe calculations
def make_chi1(phi, theta, wl, ab, c1_include):
    """
    # ##
    # Aberration function chi without defocus
    # ##
    """
    t0 = np.power(theta, 1) / 1 * (float(ab['C01a']) * np.cos(1 * phi) + float(ab['C01b']) * np.sin(1 * phi))

    if c1_include == 1:  # First and second terms
        t1 = np.power(theta, 2) / 2 * (ab['C10'] + ab['C12a'] * np.cos(2 * phi) + ab['C12b'] * np.sin(2 * phi))
    elif c1_include == 2:  # Second terms only
        t1 = np.power(theta, 2) / 2 * (ab['C12a'] * np.cos(2 * phi) + ab['C12b'] * np.sin(2 * phi))
    else:  # none for zero
        t1 = t0 * 0.

    t2 = np.power(theta, 3) / 3 * (ab['C21a'] * np.cos(1 * phi) + ab['C21b'] * np.sin(1 * phi)
                                   + ab['C23a'] * np.cos(3 * phi) + ab['C23b'] * np.sin(3 * phi))

    t3 = np.power(theta, 4) / 4 * (ab['C30']
                                   + ab['C32a'] * np.cos(2 * phi)
                                   + ab['C32b'] * np.sin(2 * phi)
                                   + ab['C34a'] * np.cos(4 * phi)
                                   + ab['C34b'] * np.sin(4 * phi))

    t4 = np.power(theta, 5) / 5 * (ab['C41a'] * np.cos(1 * phi)
                                   + ab['C41b'] * np.sin(1 * phi)
                                   + ab['C43a'] * np.cos(3 * phi)
                                   + ab['C43b'] * np.sin(3 * phi)
                                   + ab['C45a'] * np.cos(5 * phi)
                                   + ab['C45b'] * np.sin(5 * phi))

    t5 = np.power(theta, 6) / 6 * (ab['C50']
                                   + ab['C52a'] * np.cos(2 * phi)
                                   + ab['C52b'] * np.sin(2 * phi)
                                   + ab['C54a'] * np.cos(4 * phi)
                                   + ab['C54b'] * np.sin(4 * phi)
                                   + ab['C56a'] * np.cos(6 * phi)
                                   + ab['C56b'] * np.sin(6 * phi))

    chi = t0 + t1 + t2 + t3 + t4 + t5
    if 'C70' in ab:
        chi += np.power(theta, 8) / 8 * (ab['C70'])

    return chi * 2 * np.pi / wl  # np.power(theta,6)/6*(  ab['C50'] )


def probe2(ab, size_x, size_y, tags, verbose=False):
    """

    * This function creates a incident STEM probe
    * at position (0,0)
    * with parameters given in ab dictionary
    *
    * The following Aberration functions are being used:
    * 1) ddf = Cc*de/E  but not  + Cc2*(de/E)^2,
    *    Cc, Cc2 = chrom. Aber. (1st, 2nd order) [1]
    * 2) chi(qx,qy) = (2*pi/lambda)*{0.5*C1*(qx^2+qy^2)+
    *                 0.5*C12a*(qx^2-qy^2)+
    *                 C12b*qx*qy+
    *                 C21a/3*qx*(qx^2+qy^2)+
    *                 ...
    *                 +0.5*C3*(qx^2+qy^2)^2
    *                 +0.125*C5*(qx^2+qy^2)^3
    *                 ... (need to finish)
    *
    *
    *    qx = acos(kx/K), qy = acos(ky/K)
    *
    * References:
    * [1] J. Zach, M. Haider,
    *    "Correction of spherical and Chromatic Aberration
    *     in a low Voltage SEM", Optik 98 (3), 112-118 (1995)
    * [2] O.L. Krivanek, N. Delby, A.R. Lupini,
    *    "Towards sub-Angstroem Electron Beams",
    *    Ultramicroscopy 78, 1-11 (1999)
    *


    # Internally reciprocal lattice vectors in 1/nm or rad.
    # All calculations of chi in angles.
    # All aberration coefficients in nm
    """

    if 'fov' not in ab:
        if 'fov' not in tags:
            print(' need field of view in tags ')
        else:
            ab['fov'] = tags['fov']

    if 'convAngle' not in ab:
        ab['convAngle'] = 30  # in mrad

    ap_angle = ab['convAngle'] / 1000.0  # in rad

    e0 = ab['EHT'] = float(ab['EHT'])  # acceleration voltage in ev

    # defocus = ab['C10']

    if 'C01a' not in ab:
        ab['C01a'] = 0.
    if 'C01b' not in ab:
        ab['C01b'] = 0.

    if 'C50' not in ab:
        ab['C50'] = 0.
    if 'C70' not in ab:
        ab['C70'] = 0.

    if 'Cc' not in ab:
        ab['Cc'] = 1.3e6  # Cc in  nm

    def get_wl():
        h = 6.626 * 10 ** -34
        m0 = 9.109 * 10 ** -31
        ev = 1.602 * 10 ** -19 * e0
        c = 2.998 * 10 ** 8
        return h / np.sqrt(2 * m0 * ev * (1 + ev / (2 * m0 * c ** 2))) * 10 ** 9

    wl = get_wl()
    if verbose:
        print('Acceleration voltage {0:}kV => wavelength {1:.2f}pm'.format(int(e0 / 1000), wl * 1000))
    ab['wavelength'] = wl

    # Reciprocal plane in 1/nm
    dk = 1 / ab['fov']
    kx = np.array(dk * (-size_x / 2. + np.arange(size_x)))
    ky = np.array(dk * (-size_y / 2. + np.arange(size_y)))
    t_xv, t_yv = np.meshgrid(kx, ky)

    # define reciprocal plane in angles
    phi = np.arctan2(t_xv, t_yv)
    theta = np.arctan2(np.sqrt(t_xv ** 2 + t_yv ** 2), 1 / wl)

    # calculate chi but omit defocus
    chi = np.fft.ifftshift(make_chi1(phi, theta, wl, ab, 2))
    probe = np.zeros((size_x, size_y))

    # Aperture function
    mask = theta >= ap_angle

    # Calculate probe with Cc

    for i in range(len(ab['zeroLoss'])):
        df = ab['C10'] + ab['Cc'] * ab['zeroEnergy'][i] / e0
        if verbose:
            print('defocus due to Cc: {0:.2f} nm with weight {1:.2f}'.format(df, ab['zeroLoss'][i]))
        # Add defocus
        chi2 = chi + np.power(theta, 2) / 2 * df
        # Calculate exponent of - i * chi
        chi_t = np.fft.ifftshift(np.vectorize(complex)(np.cos(chi2), -np.sin(chi2)))
        # Apply aperture function
        chi_t[mask] = 0.
        # inverse fft of aberration function
        i2 = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(chi_t)))
        # add intensities
        probe = probe + np.real(i2 * np.conjugate(i2)).T * ab['zeroLoss'][i]

    ab0 = {}
    for key in ab:
        ab0[key] = 0.
    # chiIA = np.fft.fftshift(make_chi1(phi, theta, wl, ab0, 0))  # np.ones(chi2.shape)*2*np.pi/wl
    chi_i = np.ones((size_y, size_x))
    chi_i[mask] = 0.
    i2 = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(chi_i)))
    ideal = np.real(i2 * np.conjugate(i2))

    probe_f = np.fft.fft2(probe, probe.shape) + 1e-12
    ideal_f = np.fft.fft2(ideal, probe.shape)
    fourier_space_division = ideal_f / probe_f
    probe_r = (np.fft.ifft2(fourier_space_division, probe.shape))

    return probe / sum(ab['zeroLoss']), np.real(probe_r)



"""
utilities for crystal structures
"""
import numpy as np
from math import gcd, ceil, atan
from fractions import Fraction


# SIGMA_SYMBOL = u'\u03A3'
UNIMODULAR_MATRIX = np.array([np.identity(3),
                              [[1, 0, 1],
                               [0, 1, 0],
                               [0, 1, 1]],
                              [[1, 0, 1],
                               [0, 1, 0],
                               [0, 1, -1]],
                              [[1, 0, 1],
                               [0, 1, 0],
                               [-1, 1, 0]],
                              [[1, 0, 1],
                               [1, 1, 0],
                               [1, 1, 1]]])
STRUCTURE_MATRIX = np.array([np.identity(3),
                             [[0.5, -0.5, 0],
                              [0.5, 0.5, 0],
                              [0.5, 0.5, 1]],
                             [[0.5, 0.5, 0],
                              [0, 0.5, 0.5],
                              [0.5, 0, 0.5]]])


def coprime(a, b):
    """0 is coprime only with 1"""

    return gcd(a, b) in (0, 1)


def get_cubic_sigma(hkl, m, n=1):
    sqsum = np.inner(hkl, hkl)
    sigma = m*m + n*n * sqsum
    while sigma != 0 and sigma % 2 == 0:
        sigma /= 2
    return (sigma if sigma > 1 else None)


def get_cubic_theta(hkl, m, n=1):
    h, k, l = hkl
    sqsum = h*h + k*k + l*l
    assert sqsum > 0
    if m > 0:
        return 2 * atan(np.sqrt(sqsum) * n / m)
    else:
        return np.pi


def get_theta_m_n_list(hkl, sigma, verbose=False):
    if sigma == 1:
        return [(0., 0, 0)]
    thetas = []

    # From Grimmer, Acta Cryst. (1984). A40, 108-112
    #    S = m^2 + (u^2+v^2+w^2) n^2     (eq. 2)
    #    S = alpha * Sigma               (eq. 4)
    #   where alpha = 1, 2 or 4.
    # Since (u^2+v^2+w^2) n^2 > 0,
    # thus alpha * Sigma > m^2    =>   m^2 < 4 * Sigma
    max_m = int(ceil(np.sqrt(4*sigma)))

    for m in range(max_m):
        for n in range(1, max_m):
            if not coprime(m, n):
                continue
            s = get_cubic_sigma(hkl, m, n)
            if s != sigma:
                continue
            theta = get_cubic_theta(hkl, m, n)
            if verbose:
                print("m=%i n=%i" % (m, n), "%.3f" % np.degrees(theta))
            thetas.append((theta, m, n))
    return np.array(thetas)


def rodrigues(a, angle, verbose=False):
    """use Rodrigues' rotation formula to get rotation matrix"""
    a = np.array(a, dtype=float)
    a /= np.sqrt(np.inner(a, a))  # make unit vector
    # assert abs(sin_angle - sin(acos(cos_angle))) < 1e-6
    if verbose:
        print("rotation angle:", np.degrees(angle))
        print("rotation axis:", a)
    omega = np.array([[0., -a[2], a[1]],
                     [a[2], 0., -a[0]],
                     [-a[1], a[0], 0.]])
    rm = (np.identity(3) + omega * np.sin(angle) + np.dot(omega, omega) * (1 - np.cos(angle)))
    if verbose:
        print("rotation matrix:", rm)
    return rm


def is_integer(a, epsilon=1e-7):
    """return true if numpy Float array consists off all integers"""
    return (np.abs(a - np.round(a)) < epsilon).all()


def get_smallest_multiplier(a, max_n=1000):
    """return the smallest positive integer n such that matrix a multiplied
       by n is an integer matrix
    """
    for i in range(1, max_n):
        if is_integer(i*a):
            return i


def plus_minus_gen(start, end):
    """
    Generate a list of plus and minus alternating integers
    """
    for i in range(start, end):
        yield i
        yield -i


def get_csl_matrix(sigma, rotate_matrix, centering=None):
    """  Find matrix that determines the coincidence site lattice for cubic structures.

    Based on H. Grimmer et al., Acta Cryst. (1974) A30, 197
    https://doi.org/10.1107/S056773947400043X

    Parameters
    ----------
    sigma: float:
        CSL sigma
    rotate_matrix: numpy array
        rotation matrix
    centering: string
        "f" for f.c.c., "b" for b.c.c. and None for p.c.

    Returns
    -------
    matrix: numpy array
        which column vectors are the unit vectors of the CSL.
    """

    s = STRUCTURE_MATRIX[0]
    for u in UNIMODULAR_MATRIX:
        t = np.eye(3) - np.dot(np.dot(np.dot(u, np.linalg.inv(s)),  np.linalg.inv(rotate_matrix)), s)
        if abs(np.linalg.det(t)) > 1e-6:
            break
    o_lattice = np.round(np.linalg.inv(t), 12)
    n = np.round(sigma / np.linalg.det(o_lattice), 6)
    csl_matrix = o_lattice_to_csl(o_lattice, n)
    return csl_matrix


def o_lattice_to_csl(o_lattice, n):
    """
    The algorithm was borrowed from gosam project with slight changes.
    Link to the project: https://github.com/wojdyr/gosam

    There are two major steps: (1) Manipulate the columns of O-lattice to get
    an integral basis matrix for CSL: make two columns become integers and
    the remaining column can be multiplied by n whereby the determinant
    becomes sigma. (2) Simplify CSL so its vectors acquire the shortest length:
    decrease the integers in the matrix while keeping the determinant the same
    by adding other column vectors (row vectors in the following example) to a
    column vector. If after the addition or subtraction, the maximum value or
    absolute summation of added or subtracted vector is smaller than the
    original, then proceed the addition or subtraction.
    0 0 -1      0 0 -1      0 0 -1      0 0 -1      0 0 -1
    1 2 -1  ->  1 2 0   ->  1 2 0   ->  1 2 0   ->  1 2 0
    1 -3 2      1 -3 2      1 -3 1      1 -3 0      2 -1 0

    Parameters
    ----------
    o_lattice: (3x3 array)
        O-lattice in crystal coordinates
    n: int
       Number of O-lattice units per CSL unit

    Returns
    -------
    csl: 3x3 array
        CSL matrix in crystal coordinates
    """

    csl = o_lattice.copy()
    if n < 0:
        csl[0] *= -1
        n *= -1
    while True:
        m = [get_smallest_multiplier(i) for i in csl]
        m_prod = np.prod(m)
        if m_prod <= n:
            for i in range(3):
                csl[i] *= m[i]
            if m_prod < n:
                assert n % m_prod == 0
                csl[0] *= n / m_prod
            break
        else:
            changed = False
            for i in range(3):
                for j in range(3):
                    if changed or i == j or m[i] == 1 or m[j] == 1:
                        continue
                    a, b = (i, j) if m[i] <= m[j] else (j, i)
                    for k in plus_minus_gen(1, m[b]):
                        handle = csl[a] + k * csl[b]
                        if get_smallest_multiplier(handle) < m[a]:
                            csl[a] += k * csl[b]
                            changed = True
                            break
            if not changed:
                # This situation rarely happens. Not sure if this solution is
                # legit, as det not equals to sigma. E.g. Sigma 115[113]
                for i in range(3):
                    csl[i] *= m[i]
                break
    csl = csl.round().astype(int)

    # Reshape CSL
    def simplify(l1, l2):
        x = abs(l1 + l2)
        y = abs(l1)
        changed = False
        while max(x) < max(y) or (max(x) == max(y) and sum(x) < sum(y)):
            l1 += l2
            changed = True
            x = abs(l1 + l2)
            y = abs(l1)
        return changed

    while True:
        changed = False
        for i in range(3):
            for j in range(3):
                if i != j and not changed:
                    changed = simplify(csl[i], csl[j]) or simplify(csl[i], -csl[j])
                    if changed:
                        break
        if not changed:
            break
    return csl


def orthogonalize_csl(csl, axis):
    """
    (1) Set the 3rd column of csl same as the rotation axis. The algorithm was
    borrowed from gosam project with slight changes. Link to the project:
    https://github.com/wojdyr/gosam
    (2) Orthogonalize CSL, which is essentially a Gram-Schmidt process. At the
    same time, we want to make sure the column vectors of orthogonalized csl
    has the smallest value possible. That's why we compared two different ways.
    csl = [v1, v2, v3], vi is the column vector
    u1 = v3/||v3||, y2 = v1 - (v1 . u1)u1
    u2 = y2/||y2||, y3 = v3 - [(v3 . u1)u1 + (v3 . u2)u2]
    u3 = y3/||y3||
    """

    axis = np.array(axis)
    c = np.linalg.solve(csl.transpose(), axis)
    if not is_integer(c):
        mult = get_smallest_multiplier(c)
        c *= mult
    c = c.round().astype(int)
    ind = min([(i, v) for i, v in enumerate(c) if not np.allclose(v, 0)],
              key=lambda x: abs(x[1]))[0]
    if ind != 2:
        csl[ind], csl[2] = csl[2].copy(), -csl[ind]
        c[ind], c[2] = c[2], -c[ind]

    csl[2] = np.dot(c, csl)
    if c[2] < 0:
        csl[1] *= -1

    def get_integer(vec):
        # Used vec = np.array(vec, dtype=float) before, but does not work for
        # [5.00000000e-01, -5.00000000e-01,  2.22044605e-16] in Sigma3[112]
        vec = np.round(vec, 12)
        vec_sign = np.array([1 if abs(i) == i else -1 for i in vec])
        vec = list(abs(vec))
        new_vec = []
        if 0.0 in vec:
            zero_ind = vec.index(0)
            vec.pop(zero_ind)
            if 0.0 in vec:
                new_vec = [get_smallest_multiplier(vec) * i for i in vec]
            else:
                frac = Fraction(vec[0] / vec[1]).limit_denominator()
                new_vec = [frac.numerator, frac.denominator]
            new_vec.insert(zero_ind, 0)
        else:
            for i in range(len(vec) - 1):
                frac = Fraction(vec[i] / vec[i + 1]).limit_denominator()
                new_vec.extend([frac.numerator, frac.denominator])
            if new_vec[1] == new_vec[2]:
                new_vec = [new_vec[0], new_vec[1], new_vec[3]]
            else:
                new_vec = new_vec  # Todo: where is reduce_vector function
                # reduce_vector([new_vec[0] * new_vec[2], new_vec[1] * new_vec[2], new_vec[3] * new_vec[1]])
        assert is_integer(new_vec)
        return new_vec * vec_sign

    u1 = csl[2] / np.linalg.norm(csl[2])
    y2_1 = csl[1] - np.dot(csl[1], u1) * u1
    c0_1 = get_integer(y2_1)
    y2_2 = csl[0] - np.dot(csl[0], u1) * u1
    c0_2 = get_integer(y2_2)
    if sum(abs(c0_1)) > sum(abs(c0_2)):
        u2 = y2_2 / np.linalg.norm(y2_2)
        y3 = csl[1] - np.dot(csl[1], u1) * u1 - np.dot(csl[1], u2) * u2
        csl[1] = get_integer(y3)
        csl[0] = c0_2
    else:
        u2 = y2_1 / np.linalg.norm(y2_1)
        y3 = csl[0] - np.dot(csl[0], u1) * u1 - np.dot(csl[0], u2) * u2
        csl[1] = c0_1
        csl[0] = get_integer(y3)
    for i in range(3):
        for j in range(i + 1, 3):
            if not np.allclose(np.dot(csl[i], csl[j]), 0):
                raise ValueError("Non-orthogonal basis: %s" % csl)
    return csl.round().astype(int)


def find_smallest_real_multiplier(a, max_n=1000):
    """return the smallest positive real f such that matrix `a' multiplied
       by f is an integer matrix
    """

    # the smallest non-zero element|
    m = min(abs(i) for i in a if abs(i) > 1e-9)
    for i in range(1, max_n):
        t = i / float(m)
        if is_integer(t * a):
            return t
    raise ValueError("Sorry, we can't make this matrix integer:\n%s" % a)


def scale_to_integers(v):
    """scale to integer"""
    return np.array(v * find_smallest_real_multiplier(v)).round().astype(int)
